// Copyright (c) 1994, 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif
#include "config.h"
#include "EntityToolkit.H"
#include "PosixStorage.H"
#include "URLStorage.H"
#include "macros.H"
#ifdef MULTI_BYTE
#include "UTF8CodingSystem.H"
#include "Fixed2CodingSystem.H"
#include "UnicodeCodingSystem.H"
#include "EUCJPCodingSystem.H"
#include "SJISCodingSystem.H"
#include "ISO8859InputCodingSystem.H"
#endif
#include "IdentityCodingSystem.H"
#include "Vector.H"
#include "ExtendEntityManager.H"

#ifndef SGML_SEARCH_PATH_DEFAULT
#define SGML_SEARCH_PATH_DEFAULT ""
#endif

#ifdef MSDOS_FILENAMES
#define FILE_SEP ';'
#else
#define FILE_SEP ':'
#endif

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_GETTEXT
extern "C" {
extern char *dgettext(const char *, const char *);
extern char *gettext(const char *);
extern char *textdomain(const char *);
extern char *bindtextdomain(const char *, const char *);
}
#endif /* HAVE_GETTEXT */

#ifdef MULTI_BYTE
static UnivCharsetDesc::Range range = { 0, 65536, 0 };
#else
static UnivCharsetDesc::Range range = { 0, 256, 0 };
#endif

#ifdef MULTI_BYTE
static UTF8CodingSystem utf8CodingSystem;
static Fixed2CodingSystem fixed2CodingSystem;
static UnicodeCodingSystem unicodeCodingSystem;
static EUCJPCodingSystem eucjpCodingSystem;
static SJISCodingSystem sjisCodingSystem;
#endif
static IdentityCodingSystem identityCodingSystem;

static struct {
  const char *name;
  const CodingSystem *cs;
} codingSystems[] = {
#ifdef MULTI_BYTE
  { "UTF-8", &utf8CodingSystem },
  { "FIXED-2", &fixed2CodingSystem },
  { "UNICODE", &unicodeCodingSystem },
  { "EUC-JP", &eucjpCodingSystem },
  { "SJIS", &sjisCodingSystem },
#endif
  { "IDENTITY", &identityCodingSystem },
};

#ifdef MULTI_BYTE
static ISO8859InputCodingSystem iso8859_2InputCodingSystem(2);
static ISO8859InputCodingSystem iso8859_3InputCodingSystem(3);
static ISO8859InputCodingSystem iso8859_4InputCodingSystem(4);
static ISO8859InputCodingSystem iso8859_5InputCodingSystem(5);
static ISO8859InputCodingSystem iso8859_6InputCodingSystem(6);
static ISO8859InputCodingSystem iso8859_7InputCodingSystem(7);
static ISO8859InputCodingSystem iso8859_8InputCodingSystem(8);
static ISO8859InputCodingSystem iso8859_9InputCodingSystem(9);

static struct {
  const char *name;
  const InputCodingSystem *cs;
} inputCodingSystems[] = {
 { "IS8859-1", &identityCodingSystem },
 { "IS8859-2", &iso8859_2InputCodingSystem },
 { "IS8859-3", &iso8859_3InputCodingSystem },
 { "IS8859-4", &iso8859_4InputCodingSystem },
 { "IS8859-5", &iso8859_5InputCodingSystem },
 { "IS8859-6", &iso8859_6InputCodingSystem },
 { "IS8859-7", &iso8859_7InputCodingSystem },
 { "IS8859-8", &iso8859_8InputCodingSystem },
 { "IS8859-9", &iso8859_9InputCodingSystem },
};

#endif /* MULTI_BYTE */

EntityToolkit::EntityToolkit()
: storageManager_(0),
  fdStorageManager_(0),
  urlStorageManager_(0),
  systemCharset_(UnivCharsetDesc(&range, 1)),
  MessageReporter(&errorStream_)
{
  codingSystem_ = 0;
  const char *codingName = getenv("SP_BCTF");
  if (codingName)
    codingSystem_ = lookupCodingSystem(codingName);
  if (!codingSystem_ || codingSystem_->minBytesPerChar() > 1)
    codingSystem_ = &identityCodingSystem;
  errorStream_.open(cerr.rdbuf(), codingSystem_);
#ifndef MESSAGE_DOMAIN
#define MESSAGE_DOMAIN "sp"
#endif
#ifdef HAVE_GETTEXT
  const char *dir = getenv("TEXTDOMAINDIR");
  if (dir)
    bindtextdomain(MESSAGE_DOMAIN, dir);
#endif
}

const CodingSystem *EntityToolkit::lookupCodingSystem(const char *codingName)
{
  if (strlen(codingName) < 50) {
    char buf[50];
    strcpy(buf, codingName);
    for (char *p = buf; *p; p++)
      *p = toupper((unsigned char)*p);
    for (size_t i = 0; i < SIZEOF(codingSystems); i++)
      if (strcmp(buf, codingSystems[i].name) == 0)
	return codingSystems[i].cs;
  }
  return 0;
}

EntityToolkit::~EntityToolkit()
{
  delete storageManager_;
  delete fdStorageManager_;
  delete urlStorageManager_;
}

void EntityToolkit::setErrorStream(streambuf *sbuf)
{
  errorStream_.open(sbuf, codingSystem_);
}

#ifdef __GNUG__
typedef Vector<CString> Dummy_Vector_CString;
#endif

void EntityToolkit::initEntityManager()
{
  delete storageManager_;
  delete fdStorageManager_;
  delete urlStorageManager_;
  storageManager_ = new PosixStorageManager("OSFILE",
					    systemCharset_.desc(),
					    codingSystem_,
					    5);
  entityManager_
    = ExtendEntityManager::make(storageManager_, codingSystem_);
  fdStorageManager_ = new PosixFdStorageManager("OSFD",
						systemCharset_.desc());
  entityManager_->registerStorageManager(fdStorageManager_);
  urlStorageManager_ = new URLStorageManager("URL");
  entityManager_->registerStorageManager(urlStorageManager_);
  size_t i;
  for (i = 0; i < SIZEOF(codingSystems); i++)
    entityManager_->registerCodingSystem(codingSystems[i].name,
					 codingSystems[i].cs);
#ifdef MULTI_BYTE
  for (i = 0; i < SIZEOF(inputCodingSystems); i++)
    entityManager_->registerCodingSystem(inputCodingSystems[i].name,
					 inputCodingSystems[i].cs);
#endif
  for (i = 0; i < searchDirs_.length(); i++)
    storageManager_->addSearchDir(convertInput(searchDirs_[i]));
  {
    const char *e = getenv("SGML_SEARCH_PATH");
    if (!e)
      e = SGML_SEARCH_PATH_DEFAULT;
    if (*e) {
      CString str(convertInput(e));
      size_t i = 0;
      size_t start = 0;
      for (;;) {
	if (i == str.length() || str[i] == FILE_SEP) {
	  storageManager_->addSearchDir(CString(str.pointer() + start,
						i - start));
	  if (i == str.length())
	    break;
	  start = ++i;
	}
	else
	  i++;
      }
    }
  }
}

CString EntityToolkit::convertInput(const char *s)
{
  CString str(codingSystem_->convertIn(s));
  for (size_t i = 0; i < str.length(); i++)
    if (str[i] == '\n')
      str[i] = '\r';
  return str;
}

Boolean EntityToolkit::makeSystemId(int nFiles, char *const *files,
				    CString &result)
{
  Vector<CString> filenames(nFiles == 0 ? 1 : nFiles);
  int i;
  for (i = 0; i < nFiles; i++)
    filenames[i] = convertInput(strcmp(files[i], "-") == 0
				? "<FD>0"
				: files[i]);
  if (nFiles == 0)
    filenames[0] = convertInput("<FD>0");
  return entityManager_->mergeSystemIds(filenames.pointer(),
					filenames.length(),
					systemCharset_,
					*this,
					result);
}


CString EntityToolkit::messageText(const char *, const char *s)
{
#ifdef HAVE_GETTEXT
  s = dgettext(MESSAGE_DOMAIN, s);
#endif
  return codingSystem_->convertIn(s);
}

void EntityToolkit::addSearchDir(const char *s)
{
  searchDirs_.grow() = s;
}
