// Copyright (c) 1994, 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif
#include "config.h"
#include "ParserToolkit.H"
#include "StorageManager.H"
#include "SOEntityCatalog.H"
#include "ParserToolkitMessages.H"
#include "macros.H"

#ifdef MSDOS_FILENAMES
#define FILE_SEP ';'
#else
#define FILE_SEP ':'
#endif

#ifndef SGML_CATALOG_FILES_DEFAULT
#define SGML_CATALOG_FILES_DEFAULT ""
#endif /* not SGML_CATALOG_FILES_DEFAULT */

#include <string.h>
#include <stdlib.h>

ParserToolkit::ParserToolkit()
: EntityToolkit()
{
}

#ifdef __GNUG__
typedef Vector<CString> Dummy_Vector_CString;
#endif

Boolean ParserToolkit::init(int nFiles, char *const *files, unsigned flags)
{
  initEntityManager();
  CString documentSysid;
  if (!makeSystemId(nFiles, files, documentSysid))
    return 0;
  GrowableVectorD<CString> v;
  if (flags & documentFromCatalog)
    documentSysid.moveTo(v.grow());
  size_t i;
  for (i = 0; i < catalogSysids_.length(); i++)
    // filenames specified on command-line must exist
    v.grow() = convertInput(catalogSysids_[i]);
  addCatalogForDocument(documentSysid, v);
  const char *e = getenv("SGML_CATALOG_FILES");
  if (!e)
    e = SGML_CATALOG_FILES_DEFAULT;
  if (*e) {
    // FIXME Should probably convert before splitting.
    String<char> str(e, strlen(e) + 1);
    char *start = str.pointer();
    char *p = start;
    for (;;) {
      if (*p == '\0') {
	v.grow() = convertInput(start);
	break;
      }
      if (*p == FILE_SEP) {
	*p++ = '\0';
	v.grow() = convertInput(start);
	start = p;
      }
      else
	p++;
    }
  }
  SOEntityCatalog *catalog
    = SOEntityCatalog::make(v.pointer(),
			    v.length(),
			    catalogSysids_.length()
			    + (flags & documentFromCatalog) != 0,
			    systemCharset_,
			    systemCharset_,
			    entityManager().pointer(),
			    *this);
  if (flags & documentFromCatalog) {
    if (!catalog->document(systemCharset_, *this, documentSysid)) {
      message(ParserToolkitMessages::noDocumentEntry);
      return 0;
    }
  }
  // Explicit constructors are necessary for MSVC 2.1.
  parser_.init(documentSysid,
	       ResourcePointer<InputSourceOrigin>(new InputSourceOrigin),
	       ConstResourcePointer<EntityManager>(entityManager().pointer()),
	       ConstResourcePointer<EntityCatalog>(catalog),
	       systemCharset_.desc(),
	       options);
  for (i = 0; i < includes_.length(); i++)
    parser_.addInclude(convertInput(includes_[i]));
  for (i = 0; i < activeLinkTypes_.length(); i++) 
    parser_.activateLinkType(convertInput(activeLinkTypes_[i]));
  if (flags & allLinkTypesActivated)
    parser_.allLinkTypesActivated();
  return 1;
}


void ParserToolkit::addCatalogSysid(const char *s)
{
  catalogSysids_.grow() = s;
}

void ParserToolkit::addInclude(const char *s)
{
  includes_.grow() = s;
}

void ParserToolkit::activateLinkType(const char *s)
{
  activeLinkTypes_.grow() = s;
}

Boolean ParserToolkit::enableWarning(const char *s)
{
  struct {
    const char *name;
    PackedBoolean ParserOptions::*ptr;
    PackedBoolean partOfAll;
  } table[] = {
    { "mixed", &ParserOptions::warnMixedContent, 1 },
    { "should", &ParserOptions::warnShould, 1 },
    { "duplicate", &ParserOptions::warnDuplicateEntity, 0 },
    { "default", &ParserOptions::warnDefaultEntityReference, 1 },
    { "undefined", &ParserOptions::warnUndefinedElement, 1 },
    { "sgmldecl", &ParserOptions::warnSgmlDecl, 1 },
    { "unclosed", &ParserOptions::warnUnclosedTag, 1 },
    { "empty", &ParserOptions::warnEmptyTag, 1 },
    { "net", &ParserOptions::warnNet, 0 },
    { "unused-map", &ParserOptions::warnUnusedMap, 1 },
    { "unused-param", &ParserOptions::warnUnusedParam, 1 },
    { "idref", &ParserOptions::errorIdref, 0 },
    { "significant", &ParserOptions::errorSignificant, 0 },
  };
  PackedBoolean val = 1;
  if (strncmp(s, "no-", 3) == 0) {
    s += 3;
    val = 0;
  }
  if (strcmp(s, "all") == 0) {
    for (size_t i = 0; i < SIZEOF(table); i++)
      if (table[i].partOfAll) {
	// Use parentheses to work around Watcom 10.0a bug.
	(options.*(table[i].ptr)) = val;
      }
    return 1;
  }
  for (size_t i = 0; i < SIZEOF(table); i++)
    if (strcmp(s, table[i].name) == 0) {
      // Use parentheses to work around Watcom 10.0a bug.
      (options.*(table[i].ptr)) = val;
      return 1;
    }
  if (strcmp(s, "min-tag") == 0) {
    options.warnUnclosedTag = val;
    options.warnEmptyTag = val;
    options.warnNet = val;
    return 1;
  }
  return 0;
}

void ParserToolkit::addCatalogForDocument(const CString &sysid,
					  GrowableVectorD<CString> &catalog)
{
  ParsedSystemId v;
  if (!entityManager()->parseSystemId(sysid, systemCharset_, 0, 0, *this, v))
    return;
  size_t catalogInitLength = catalog.length();
  for (size_t i = 0; i < v.length(); i++) {
    ParsedSystemId catalogId;
    StorageObjectSpec &spec = catalogId.grow();
    spec.storageManager = v[i].storageManager;
    spec.codingSystemName = v[i].codingSystemName;
    spec.specId = spec.storageManager->idCharset().execToDesc("catalog");
    spec.storageManager->resolveRelative(v[i].specId, spec.specId, 1);
    spec.baseId = v[i].baseId;
    spec.records = v[i].records;
    CString tem;
    catalogId.unparse(systemCharset_, tem);
    for (size_t j = catalogInitLength; j < catalog.length(); j++)
      if (tem == catalog[j]) {
	tem.clear();
	break;
      }
    if (tem.length() > 0)
      tem.moveTo(catalog.grow());
  }
}
