// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Location_INCLUDED
#define Location_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "types.H"
#include "Boolean.H"
#include "ResourcePointer.H"
#include "Resource.H"
#include "Boolean.H"
#include "GrowableVector.H"
#include "Owner.H"
#include "CString.H"
#include "rtti.H"

class ExternalInfo;
class EntityOrigin;
class InputSourceOrigin;
class Entity;
class Location;
class Markup;
class Text;
class NamedCharRef;

class Origin : public Resource {
public:
  virtual ~Origin();
  virtual const EntityOrigin *asEntityOrigin() const;
  virtual const InputSourceOrigin *asInputSourceOrigin() const;
  virtual const Location &parent() const = 0;
  virtual Index refLength() const;
  virtual Boolean origChar(Char &) const;
  virtual Boolean inBracketedTextOpenDelim() const;
  virtual Boolean inBracketedTextCloseDelim() const;
  virtual Boolean isNumericCharRef(const Markup *&markup) const;
  virtual Boolean isNamedCharRef(Index ind, NamedCharRef &ref) const;
};

class Location {
public:
  Location();
  Location(Origin *, Index);
  Location(ConstResourcePointer<Origin>, Index);
  void operator+=(Index i) { index_ += i; }
  void operator-=(Index i) { index_ -= i; }
  Index index() const { return index_; }
  const ConstResourcePointer<Origin> &origin() const { return origin_; }
  void clear() { origin_.clear(); }
private:
  ConstResourcePointer<Origin> origin_;
  Index index_;
};

class ExternalInfo {
  RTTI_CLASS
public:
  virtual ~ExternalInfo();
};

class NamedCharRef {
public:
  enum RefEndType {
    endOmitted,
    endRE,
    endRefc
    };
  NamedCharRef();
  NamedCharRef(Index, RefEndType, const CString &);
  Index refStartIndex() const;
  RefEndType refEndType() const;
  const CString &origName() const;
  void set(Index, RefEndType, const Char *, size_t);
private:
  Index refStartIndex_;
  RefEndType refEndType_;
  CString origName_;
};

struct InputSourceOriginNamedCharRef {
  Index replacementIndex;
  size_t origNameOffset;
  Index refStartIndex;
  NamedCharRef::RefEndType refEndType;
};

class InputSourceOrigin : public Origin {
public:
  InputSourceOrigin();
  InputSourceOrigin(const Location &refLocation);
  const Location &parent() const;
  const ExternalInfo *externalInfo() const;
  Offset startOffset(Index ind) const;
  void noteCharRef(Index replacementIndex, const NamedCharRef &);
  Boolean isNamedCharRef(Index ind, NamedCharRef &ref) const;
  void setExternalInfo(ExternalInfo *);
  virtual Boolean defLocation(Offset off, Location &) const;
  virtual InputSourceOrigin *copy() const;
  const InputSourceOrigin *asInputSourceOrigin() const;
  virtual const CString *entityName() const;
private:
  InputSourceOrigin(const InputSourceOrigin &); // undefined
  void operator=(const InputSourceOrigin &);	// undefined
  size_t nPrecedingCharRefs(Index ind) const;
  GrowableVector<InputSourceOriginNamedCharRef> charRefs_;
  CString charRefOrigNames_;
  Owner<ExternalInfo> externalInfo_; // 0 for internal entities
  Location refLocation_;	// where referenced from
};

// a delimiter specified in bracketed text

class BracketOrigin : public Origin {
public:
  enum Position { open, close };
  BracketOrigin(const Location &, Position);
  const Location &parent() const;
  Boolean inBracketedTextOpenDelim() const;
  Boolean inBracketedTextCloseDelim() const;
private:
  Position pos_;
  Location loc_;
};

class ReplacementOrigin : public Origin {
public:
  ReplacementOrigin(const Location &, Char origChar);
  const Location &parent() const;
  Boolean origChar(Char &) const;
private:
  Location loc_;
  Char origChar_;
};

inline
Index NamedCharRef::refStartIndex() const
{
  return refStartIndex_;
}

inline
NamedCharRef::RefEndType NamedCharRef::refEndType() const
{
  return refEndType_;
}

inline
const CString &NamedCharRef::origName() const
{
  return origName_;
}

inline
const ExternalInfo *InputSourceOrigin::externalInfo() const
{
  return externalInfo_.pointer();
}

#endif /* not Location_INCLUDED */
