// Copyright (c) 1994, 1995 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "SgmlParser.H"
#include "Event.H"
#include "MessageEventHandler.H"
#include "SgmlsEventHandler.H"
#include "RastEventHandler.H"
#include "OutputCharStream.H"
#include "ParserToolkit.H"
#include "Boolean.H"
#include "programName.H"
#include "getopt.h"
#include "macros.H"
#include "NsgmlsMessages.H"
#include "MessageArg.H"
#include "ErrnoMessageArg.H"

#include "outOfMemory.H"
#include "xnew.h"

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef HAVE_LOCALE
#include <locale.h>
#endif

inline
void usage(Messenger &mgr)
{
  mgr.message(NsgmlsMessages::usageInfo);
}

struct {
  const char *name;
  unsigned flag;
} outputOptions[] = {
  { "line", SgmlsEventHandler::outputLine },
  { "entity", SgmlsEventHandler::outputEntity },
  { "id", SgmlsEventHandler::outputId },
  { "included", SgmlsEventHandler::outputIncluded },
};

static int parseProlog(ParserToolkit &kit, unsigned flags);
static int validate(ParserToolkit &kit, unsigned flags);
static int outputSgmls(ParserToolkit &kit, streambuf *,
		       const OutputCodingSystem *,
		       unsigned outputFlags);
static int outputRast(ParserToolkit &kit, const char *filename,
		      const OutputCodingSystem *);

int main(int argc, char **argv)
{
  set_new_handler(outOfMemory);
#ifdef HAVE_LOCALE
  setlocale(LC_ALL, "");
#endif
  programName = argv[0];
  int opt;
  Boolean suppressOutput = 0;
  ParserToolkit kit;
  Boolean prologOnly = 0;
  unsigned outputFlags = 0;
  unsigned initFlags = 0;;
  const char *errorFile = 0;
  const char *rastFile = 0;
  const CodingSystem *outputCodingSystem = 0;
  char arg[2];

  opterr = 0;			// no messages from getopt
  // Colon at beginning is Posix.2ism that says to return : rather than ? for
  // missing option argument.
  while ((opt = getopt(argc, argv, ":a:b:c:dCD:ef:gi:lm:no:prst:uvw:T"))
	 >= 0) {
    switch (opt) {
    case 'a':
      // activate link
      kit.activateLinkType(optarg);
      break;
    case 'b':
      outputCodingSystem = kit.lookupCodingSystem(optarg);
      if (!outputCodingSystem)
	kit.message(NsgmlsMessages::unknownBctf,
		       StringMessageArg(kit.codingSystem()
					->convertIn(optarg)));
      break;
    case 'c':
      // output RACT file
      kit.message(NsgmlsMessages::noRactOption);
      break;
    case 'C':
      initFlags |= ParserToolkit::documentFromCatalog;
      break;
    case 'd':
      // warn about duplicate entity declarations
      kit.options.warnDuplicateEntity = 1;
      break;
    case 'D':
      kit.addSearchDir(optarg);
      break;
    case 'e':
      // describe open entities in error messages
      kit.addOption(MessageReporter::openEntities);
      break;
    case 'f':
      errorFile = optarg;
      break;
    case 'g':
      // show gis of open elements in error messages
      kit.addOption(MessageReporter::openElements);
      break;
    case 'i':
      // pretend that optarg is defined as INCLUDE
      kit.addInclude(optarg);
      break;
    case 'l':
      // output L commands
      outputFlags |= SgmlsEventHandler::outputLine;
      break;
    case 'm':
      kit.addCatalogSysid(optarg);
      break;
    case 'n':
      kit.options.generateNotationSystemId = 1;
      break;
    case 'o':
      {
	Boolean found = 0;
	for (size_t i = 0; i < SIZEOF(outputOptions); i++)
	  if (strcmp(optarg, outputOptions[i].name) == 0) {
	    outputFlags |= outputOptions[i].flag;
	    found = 1;
	    break;
	  }
	if (!found)
	  kit.message(NsgmlsMessages::unknownOutputOption,
		       StringMessageArg(kit.codingSystem()
					->convertIn(optarg)));
      }
      break;
    case 'p':
      prologOnly = 1;
      break;
    case 'r':
      // warn about defaulted entity reference
      kit.options.warnDefaultEntityReference = 1;
      break;
    case 's':
      suppressOutput = 1;
      break;
    case 't':
      rastFile = optarg;
      break;
    case 'T':
      kit.addOption(MessageReporter::messageTags);
      break;
    case 'u':
      // warn about undefined elements
      kit.options.warnUndefinedElement = 1;
      break;
    case 'v':
      // print the version number
      {
	extern const char *versionString;
	// FIXME perhaps we should have a constant argument instead
	kit.message(NsgmlsMessages::versionInfo,
		     StringMessageArg(kit.codingSystem()
				      ->convertIn(versionString)));
      }
      break;
    case 'w':
      if (!kit.enableWarning(optarg))
	kit.message(NsgmlsMessages::unknownWarning,
		     StringMessageArg(kit.codingSystem()
				      ->convertIn(optarg)));
      break;
    case ':':
      arg[0] = optopt;
      arg[1] = '\0';
      kit.message(NsgmlsMessages::missingOptionArgError,
		   StringMessageArg(kit.codingSystem()->convertIn(arg)));
      usage(kit);
      return 1;
    case '?':
      arg[0] = optopt;
      arg[1] = '\0';
      kit.message(NsgmlsMessages::invalidOptionError,
		   StringMessageArg(kit.codingSystem()->convertIn(arg)));
      usage(kit);
      return 1;
    default:
      CANNOT_HAPPEN();
    }
  }
#ifdef __GNUG__
  // cout is a performance disaster in libg++ unless we do this.
  ios::sync_with_stdio(0);
#endif
  if (errorFile) {
    static filebuf file;
    if (!file.open(errorFile, ios::out|ios::trunc)) {
      kit.message(NsgmlsMessages::cannotOpenOutputError,
		   StringMessageArg(kit.codingSystem()
				    ->convertIn(errorFile)),
		   ErrnoMessageArg(errno));
      exit(1);
    }
    kit.setErrorStream(&file);
  }
  if (!rastFile)
    initFlags |= ParserToolkit::allLinkTypesActivated;
  if (!outputCodingSystem)
    outputCodingSystem = kit.codingSystem();
  // could have option to parse all declared subdocuments
  if (!kit.init(argc - optind, argv + optind, initFlags))
    return 1;
  if (prologOnly)
    return parseProlog(kit, 0);
  else if (rastFile)
    return outputRast(kit, rastFile, outputCodingSystem);
  else if (suppressOutput)
    return validate(kit, MessageEventHandler::parseSubdocs);
  else 
    return outputSgmls(kit, cout.rdbuf(), outputCodingSystem, outputFlags);
}

static
int parseProlog(ParserToolkit &kit, unsigned flags)
{
  SgmlParser &parser = kit.parser();
  MessageEventHandler handler(&parser, &kit, flags);
  Event *event;
  while ((event = parser.nextEvent()) != 0
	 && event->type() != Event::endProlog)
    event->handle(handler);
  return handler.errorCount() > 0;
}

static
int validate(ParserToolkit &kit, unsigned flags)
{
  SgmlParser &parser = kit.parser();
  MessageEventHandler handler(&parser, &kit, flags);
  parser.parseAll(handler);
  return handler.errorCount() > 0;
}

static
int outputSgmls(ParserToolkit &kit, streambuf *stream,
		const OutputCodingSystem *ocs, unsigned outputFlags)
{
  IosOutputCharStream os(stream, ocs);
  SgmlParser &parser = kit.parser();
  SgmlsEventHandler handler(&parser, &os, &kit, outputFlags);
  parser.parseAll(handler);
  handler.endDocument();
  return handler.errorCount() > 0;
}

class XRastEventHandler : public RastEventHandler {
public:
  XRastEventHandler(SgmlParser *,
		    const char *filename,
		    const OutputCodingSystem *,
		    const InputCodingSystem *,
		    MessageReporter *reporter);
  void message(MessageEvent *);
  void truncateOutput();
private:
  MessageReporter *reporter_;
  // file_ must come before os_ so it gets inited first
  filebuf file_;
  IosOutputCharStream os_;
  const char *filenameBytes_;
  const InputCodingSystem *filenameCodingSystem_;
};

XRastEventHandler::XRastEventHandler(SgmlParser *parser,
				     const char *filenameBytes,
				     const OutputCodingSystem *codingSystem,
				     const InputCodingSystem *filenameCodingSystem,
				     MessageReporter *reporter)
: RastEventHandler(parser, reporter),
  reporter_(reporter),
  filenameBytes_(filenameBytes),
  filenameCodingSystem_(filenameCodingSystem)
{
  errno = 0;
  if (!file_.open(filenameBytes, ios::out|ios::trunc)) {
    reporter->message(NsgmlsMessages::cannotOpenOutputError,
		      StringMessageArg(filenameCodingSystem_
				       ->convertIn(filenameBytes_)),
		      ErrnoMessageArg(errno));
    exit(1);
  }
  os_.open(&file_, codingSystem);
  setOutputStream(&os_);
}

void XRastEventHandler::truncateOutput()
{
  os_.flush();
  errno = 0;
  if (!file_.close())
    reporter_->message(NsgmlsMessages::closeOutputError,
		       StringMessageArg(filenameCodingSystem_
					->convertIn(filenameBytes_)),
		       ErrnoMessageArg(errno));
  errno = 0;
  if (!file_.open(filenameBytes_, ios::out|ios::trunc)) {
    reporter_->message(NsgmlsMessages::cannotOpenOutputError,
		       StringMessageArg(filenameCodingSystem_
					->convertIn(filenameBytes_)),
		       ErrnoMessageArg(errno));
    exit(1);
  }
}

void XRastEventHandler::message(MessageEvent *event)
{
  reporter_->reportMessage(event->message());
  RastEventHandler::message(event);
}

static
int outputRast(ParserToolkit &kit, const char *filename,
	       const OutputCodingSystem *ocs)
{
  SgmlParser &parser = kit.parser();
  XRastEventHandler handler(&parser, filename, ocs, kit.codingSystem(), &kit);
  parser.parseAll(handler);
  handler.end();
  return handler.hadError();
}
