// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Dtd_INCLUDED
#define Dtd_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "NamedTable.H"
#include "NamedResourceTable.H"
#include "ElementType.H"
#include "Notation.H"
#include "Entity.H"
#include "ShortReferenceMap.H"
#include "Resource.H"
#include "CString.H"
#include "StringResource.H"
#include "Boolean.H"
#include "GrowableVectorD.H"
#include "HashTable.H"

class Syntax;
class ParserState;

class Dtd : public Resource {
public:
  typedef NamedTableIter<ElementType> ElementTypeIter;
  typedef ConstNamedTableIter<ElementType> ConstElementTypeIter;
  typedef NamedTableIter<ShortReferenceMap> ShortReferenceMapIter;
  typedef ConstNamedResourceTableIter<Notation> ConstNotationIter;
  typedef ConstNamedResourceTableIter<Entity> ConstEntityIter;
  typedef NamedResourceTableIter<Entity> EntityIter;
  Dtd(const CString &name, Boolean isBase);
  ConstResourcePointer<Entity> lookupEntity(Boolean isParameter,
					    const CString &) const;
  ResourcePointer<Entity> lookupEntity(Boolean isParameter, const CString &);
  ResourcePointer<Entity> insertEntity(const ResourcePointer<Entity> &);
  ResourcePointer<Entity> removeEntity(Boolean isParameter,
				       const CString &);
  ConstResourcePointer<Entity> lookupDefaultedEntity(const CString &) const;
  ResourcePointer<Entity> insertDefaultedEntity(const ResourcePointer<Entity> &);
  ResourcePointer<Entity> removeDefaultedEntity(const CString &);
  ConstEntityIter generalEntityIter() const;
  EntityIter generalEntityIter();
  ConstEntityIter parameterEntityIter() const;
  EntityIter parameterEntityIter();
  ConstEntityIter defaultedEntityIter() const;

  ConstResourcePointer<Entity> defaultEntity() const;
  void setDefaultEntity(const ResourcePointer<Entity> &, ParserState &);
  const ConstResourcePointer<StringResource<Char> > &namePointer() const;
  const CString &name() const;

  const ElementType *lookupElementType(const CString &) const;
  ElementType *lookupElementType(const CString &);
  ElementType *insertElementType(ElementType *);
  size_t nElementTypeIndex() const;
  ConstElementTypeIter elementTypeIter() const;
  ElementTypeIter elementTypeIter();

  const RankStem *lookupRankStem(const CString &) const;
  RankStem *lookupRankStem(const CString &);
  RankStem *insertRankStem(RankStem *);
  size_t nRankStem() const;

  const ShortReferenceMap *lookupShortReferenceMap(const CString &) const;
  ShortReferenceMap *lookupShortReferenceMap(const CString &);
  ShortReferenceMap *insertShortReferenceMap(ShortReferenceMap *);
  ShortReferenceMapIter shortReferenceMapIter();

  Boolean shortrefIndex(const CString &, const Syntax &, size_t &index);
  size_t nShortref() const;
  const CString &shortref(size_t i) const;
  void addNeededShortref(const CString &);

  ConstResourcePointer<Notation> lookupNotation(const CString &) const;
  ResourcePointer<Notation> lookupNotation(const CString &);
  ResourcePointer<Notation> insertNotation(const ResourcePointer<Notation> &);
  ConstNotationIter notationIter() const;

  size_t allocCurrentAttributeIndex();
  size_t nCurrentAttribute() const;
  size_t allocElementDefinitionIndex();
  size_t nElementDefinition() const;
  size_t allocAttributeDefinitionListIndex();
  size_t nAttributeDefinitionList() const;
  const ElementType *documentElementType() const;
  Boolean isBase() const;
private:
  Dtd(const Dtd &);		// undefined
  void operator=(const Dtd &);	// undefined
  NamedResourceTable<Entity> generalEntityTable_;
  NamedResourceTable<Entity> parameterEntityTable_;
  NamedResourceTable<Entity> defaultedEntityTable_;
  ConstResourcePointer<Entity> defaultEntity_;
  ConstResourcePointer<StringResource<Char> > name_;
  NamedTable<ElementType> elementTypeTable_;
  NamedTable<RankStem> rankStemTable_;
  NamedTable<ShortReferenceMap> shortReferenceMapTable_;
  NamedResourceTable<Notation> notationTable_;
  size_t nCurrentAttribute_;
  size_t nElementDefinition_;
  size_t nAttributeDefinitionList_;
  ElementType *documentElementType_;
  GrowableVectorD<CString> shortrefs_;
  HashTable<CString,int> shortrefTable_;
  Boolean isBase_;
};

inline
ConstResourcePointer<Entity> Dtd::lookupEntity(Boolean isParameter,
					       const CString &name) const
{
  return (isParameter
	  ? &parameterEntityTable_
	  : &generalEntityTable_)->clookup(name);
}

inline
ResourcePointer<Entity> Dtd::lookupEntity(Boolean isParameter,
					  const CString &name)
{
  return (isParameter
	  ? &parameterEntityTable_
	  : &generalEntityTable_)->lookup(name);
}

inline
ResourcePointer<Entity>
Dtd::insertEntity(const ResourcePointer<Entity> &entity)
{
  return (entity->declType() == Entity::parameterEntity
	  ? &parameterEntityTable_
	  : &generalEntityTable_)->insert(entity);
}

inline
ResourcePointer<Entity> Dtd::removeEntity(Boolean isParameter,
					  const CString &name)
{
  return (isParameter
	  ? &parameterEntityTable_
	  : &generalEntityTable_)->remove(name);
}

inline
ConstResourcePointer<Entity>
Dtd::lookupDefaultedEntity(const CString &name) const
{
  return defaultedEntityTable_.clookup(name);
}

inline
ResourcePointer<Entity>
Dtd::insertDefaultedEntity(const ResourcePointer<Entity> &entity)
{
  return defaultedEntityTable_.insert(entity);
}

inline
ResourcePointer<Entity> Dtd::removeDefaultedEntity(const CString &name)
{
  return defaultedEntityTable_.remove(name);
}

inline
Dtd::ConstEntityIter Dtd::generalEntityIter() const
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return ConstNamedResourceTableIter<Entity>(generalEntityTable_);
}

inline
Dtd::EntityIter Dtd::generalEntityIter()
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return NamedResourceTableIter<Entity>(generalEntityTable_);
}

inline
Dtd::ConstEntityIter Dtd::parameterEntityIter() const
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return ConstNamedResourceTableIter<Entity>(parameterEntityTable_);
}

inline
Dtd::EntityIter Dtd::parameterEntityIter()
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return NamedResourceTableIter<Entity>(parameterEntityTable_);
}

inline
Dtd::ConstEntityIter Dtd::defaultedEntityIter() const
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return ConstNamedResourceTableIter<Entity>(defaultedEntityTable_);
}

inline
ConstResourcePointer<Entity> Dtd::defaultEntity() const
{
  return defaultEntity_;
}

inline
const ConstResourcePointer<StringResource<Char> > &Dtd::namePointer() const
{
  return name_;
}

inline
const CString &Dtd::name() const
{
  return *name_;
}

inline
size_t Dtd::allocCurrentAttributeIndex()
{
  return nCurrentAttribute_++;
}

inline
size_t Dtd::nCurrentAttribute() const
{
  return nCurrentAttribute_;
}

inline
size_t Dtd::allocElementDefinitionIndex()
{
  return nElementDefinition_++;
}

inline
size_t Dtd::nElementDefinition() const
{
  return nElementDefinition_;
}

inline
size_t Dtd::allocAttributeDefinitionListIndex()
{
  return nAttributeDefinitionList_++;
}

inline
size_t Dtd::nAttributeDefinitionList() const
{
  return nAttributeDefinitionList_;
}

inline
const ElementType *Dtd::lookupElementType(const CString &name) const
{
  return elementTypeTable_.lookup(name);
}

inline
ElementType *Dtd::lookupElementType(const CString &name)
{
  return elementTypeTable_.lookup(name);
}

inline
ElementType *Dtd::insertElementType(ElementType *e)
{
  return elementTypeTable_.insert(e);
}

inline
Dtd::ElementTypeIter Dtd::elementTypeIter()
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return NamedTableIter<ElementType>(elementTypeTable_);
}

inline
Dtd::ConstElementTypeIter Dtd::elementTypeIter() const
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return ConstNamedTableIter<ElementType>(elementTypeTable_);
}

inline
size_t Dtd::nElementTypeIndex() const
{
  // index 0 is reserved for #pcdata
  return 1 + elementTypeTable_.count();
}

inline
const RankStem *Dtd::lookupRankStem(const CString &name) const
{
  return rankStemTable_.lookup(name);
}

inline
RankStem *Dtd::lookupRankStem(const CString &name)
{
  return rankStemTable_.lookup(name);
}

inline
RankStem *Dtd::insertRankStem(RankStem *e)
{
  return rankStemTable_.insert(e);
}

inline
size_t Dtd::nRankStem() const
{
  return rankStemTable_.count();
}

inline
ConstResourcePointer<Notation> Dtd::lookupNotation(const CString &name) const
{
  return notationTable_.clookup(name);
}

inline
ResourcePointer<Notation> Dtd::lookupNotation(const CString &name)
{
  return notationTable_.lookup(name);
}

inline
ResourcePointer<Notation> Dtd::insertNotation(const ResourcePointer<Notation> &nt)
{
  return notationTable_.insert(nt);
}

inline
Dtd::ConstNotationIter Dtd::notationIter() const
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return ConstNamedResourceTableIter<Notation>(notationTable_);
}

inline
const ElementType *Dtd::documentElementType() const
{
  return documentElementType_;
}

inline
const ShortReferenceMap *Dtd::lookupShortReferenceMap(const CString &name) const
{
  return shortReferenceMapTable_.lookup(name);
}

inline
ShortReferenceMap *Dtd::lookupShortReferenceMap(const CString &name)
{
  return shortReferenceMapTable_.lookup(name);
}

inline
ShortReferenceMap *Dtd::insertShortReferenceMap(ShortReferenceMap *map)
{
  return shortReferenceMapTable_.insert(map);
}

inline
Dtd::ShortReferenceMapIter Dtd::shortReferenceMapIter()
{
  // Avoid use of typedef to work around MSVC 2.0 bug.
  return NamedTableIter<ShortReferenceMap>(shortReferenceMapTable_);
}

inline
Boolean Dtd::isBase() const
{
  return isBase_;
}

inline
size_t Dtd::nShortref() const
{
  return shortrefs_.length();
}

inline
const CString &Dtd::shortref(size_t i) const
{
  return shortrefs_[i];
}

#endif /* not Dtd_INCLUDED */
