// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Entity_INCLUDED
#define Entity_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "types.H"
#include "CString.H"
#include "NamedResource.H"
#include "Location.H"
#include "Owner.H"
#include "Attribute.H"
#include "ExternalId.H"
#include "Text.H"
#include "SubstTable.H"
#include "StringResource.H"
#include "Allocator.H"
#include "EntityDecl.H"

class Messenger;
class InputSource;
class EntityOrigin;
class ParserState;
class ExternalEntity;
class ExternalDataEntity;
class SubdocEntity;
class InternalEntity;
class Notation;

class Entity : public EntityDecl {
public:
  Entity(const CString &name, DeclType declType, DataType dataType,
	 const Location &defLocation);
  // reference in a literal
  virtual void litReference(Text &, ParserState &,
			    const ResourcePointer<EntityOrigin> &,
			    Boolean squeezeSpaces)
    const;
  // reference in a declaration
  virtual void declReference(ParserState &,
			     const ResourcePointer<EntityOrigin> &)
    const;
  // reference in a declaration subset
  virtual void dsReference(ParserState &,
			   const ResourcePointer<EntityOrigin> &)
    const;
  // reference in content
  virtual void contentReference(ParserState &,
				const ResourcePointer<EntityOrigin> &)
    const;
  // reference in rcdata
  virtual void rcdataReference(ParserState &,
			       const ResourcePointer<EntityOrigin> &)
    const;
  // for entity name attribute checking
  virtual Boolean isDataOrSubdoc() const;
  // for determining whether we need to validate as character data
  virtual Boolean isCharacterData() const;
  virtual const ExternalDataEntity *asExternalDataEntity() const;
  virtual const SubdocEntity *asSubdocEntity() const;
  virtual const InternalEntity *asInternalEntity() const;
  virtual const ExternalEntity *asExternalEntity() const;
  // needed for default entity
  virtual Entity *copy() const = 0;
  virtual void generateSystemId(ParserState &);
  void setUsed();
  Boolean used() const;
protected:
  static void checkEntlvl(ParserState &);
  Boolean checkNotOpen(ParserState &) const;
private:
  virtual void normalReference(ParserState &,
			       const ResourcePointer<EntityOrigin> &,
			       Boolean generateEvent) const = 0;
  Location defLocation_;
  PackedBoolean used_;
};

class InternalEntity : public Entity {
public:
  InternalEntity(const CString &, DeclType declType, DataType dataType,
		 const Location &, Text &);
  const CString &string() const;
  const Text &text() const;
  const InternalEntity *asInternalEntity() const;
protected:
  Text text_;
};

class PiEntity : public InternalEntity {
public:
  PiEntity(const CString &, DeclType, const Location &, Text &);
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean) const;
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean) const;
  void declReference(ParserState &,
		     const ResourcePointer<EntityOrigin> &) const;
  void rcdataReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &) const;
  Entity *copy() const;
};

class InternalDataEntity : public InternalEntity {
public:
  InternalDataEntity(const CString &, DataType, const Location &, Text &);
  void declReference(ParserState &,
		     const ResourcePointer<EntityOrigin> &) const;
  Boolean isDataOrSubdoc() const;
protected:
  void addChars(Text &text, Location &loc, Boolean squeeze, Char space) const;
};

class InternalCdataEntity : public InternalDataEntity {
public:
  InternalCdataEntity(const CString &, const Location &, Text &);
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean) const;
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean) const;
  Entity *copy() const;
  Boolean isCharacterData() const;
};

class InternalSdataEntity : public InternalDataEntity {
public:
  InternalSdataEntity(const CString &, const Location &, Text &);
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean) const;
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean) const;
  Entity *copy() const;
  Boolean isCharacterData() const;
};

class InternalTextEntity : public InternalEntity {
public:
  enum Bracketed {
    none,
    starttag,
    endtag,
    ms,
    md
    };
  InternalTextEntity(const CString &, DeclType, const Location &, Text &,
		     Bracketed);
  Entity *copy() const;
private:
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean) const;
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean) const;
  Bracketed bracketed_;
};

class ExternalEntity : public Entity {
public:
  ExternalEntity(const CString &, DeclType, DataType, const Location &,
		 ExternalId &);
  const ExternalId &externalId() const;
  const ExternalEntity *asExternalEntity() const;
  void generateSystemId(ParserState &);
  const CString *systemIdPointer() const;
  const CString *publicIdPointer() const;
private:
  ExternalId externalId_;
};

class ExternalTextEntity : public ExternalEntity {
public:
  ExternalTextEntity(const CString &, DeclType, const Location &,
		     ExternalId &);
  Entity *copy() const;
private:
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean) const;
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean) const;
};

class ExternalNonTextEntity : public ExternalEntity {
public:
  ExternalNonTextEntity(const CString &, DataType,
			const Location &, ExternalId &);
  Boolean isDataOrSubdoc() const;
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean) const;
  void rcdataReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &) const;
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean) const;
  Boolean isCharacterData() const;
};

class ExternalDataEntity : public ExternalNonTextEntity {
public:
  ExternalDataEntity(const CString &, DataType, const Location &,
		     ExternalId &, const ConstResourcePointer<Notation> &,
		     AttributeList &);
  const AttributeList &attributes() const;
  const Notation *notation() const;
  const ExternalDataEntity *asExternalDataEntity() const;
  Entity *copy() const;
  void contentReference(ParserState &,
			const ResourcePointer<EntityOrigin> &) const;
  void setAttributes(AttributeList &);
private:
  ConstResourcePointer<Notation> notation_;
  AttributeList attributes_;
};

class SubdocEntity : public ExternalNonTextEntity {
public:
  SubdocEntity(const CString &, const Location &, ExternalId &);
  const SubdocEntity *asSubdocEntity() const;
  Entity *copy() const;
  void contentReference(ParserState &,
			const ResourcePointer<EntityOrigin> &) const;
private:
};

class IgnoredEntity : public Entity {
public:
  IgnoredEntity(const CString &, DeclType declType);
  Entity *copy() const;
  void litReference(Text &, ParserState &,
		    const ResourcePointer<EntityOrigin> &,
		    Boolean squeezeSpaces) const;
  void declReference(ParserState &,
		     const ResourcePointer<EntityOrigin> &) const;
private:
  void normalReference(ParserState &,
		       const ResourcePointer<EntityOrigin> &,
		       Boolean generateEvent) const;
};

class EntityOrigin : public InputSourceOrigin {
public:
  void *operator new(size_t sz, Allocator &alloc) {
    return alloc.alloc(sz);
  }
  void *operator new(size_t sz) {
    return Allocator::allocSimple(sz);
  }
  void operator delete(void *p) {
    Allocator::free(p);
  }
  EntityOrigin();
  EntityOrigin(const Location &refLocation);
  EntityOrigin(const ConstResourcePointer<Entity> &);
  EntityOrigin(const ConstResourcePointer<Entity> &,
	       const Location &refLocation);
  EntityOrigin(const ConstResourcePointer<Entity> &,
	       const Location &refLocation, Index refLength,
	       Owner<Markup> &markup);
  ~EntityOrigin();
  InputSourceOrigin *copy() const;
  const ConstResourcePointer<Entity> &entity() const { return entity_; }
  const CString *entityName() const;
  const EntityOrigin *asEntityOrigin() const;
  Boolean defLocation(Offset off, Location &loc) const;
  Index refLength() const;
  const Markup *markup() const;
private:
  EntityOrigin(const EntityOrigin &); // undefined
  void operator=(const EntityOrigin &);	// undefined
  ConstResourcePointer<Entity> entity_;	// 0 for document entity
  // total length of reference
  // (characters that were replaced by the entity)
  Index refLength_;
  Owner<Markup> markup_;
};

inline
Boolean Entity::used() const
{
  return used_;
}

inline
void Entity::setUsed()
{
  used_ = 1;
}

inline
const CString &InternalEntity::string() const
{
  return text_.string();
}

inline
const Text &InternalEntity::text() const
{
  return text_;
}

inline
const ExternalId &ExternalEntity::externalId() const
{
  return externalId_;
}

inline
const AttributeList &ExternalDataEntity::attributes() const
{
  return attributes_;
}

inline
const Notation *ExternalDataEntity::notation() const
{
  return notation_.pointer();
}

inline
const Markup *EntityOrigin::markup() const
{
  return markup_.pointer();
}

#endif /* not Entity_INCLUDED */
