// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif
#include "config.h"
#include "Event.H"
#include "Entity.H"
#include "Attribute.H"
#include "EventQueue.H"

Event::~Event()
{
}

void Event::copyData()
{
}

LocatedEvent::LocatedEvent(Type type, const Location &location)
: location_(location), Event(type)
{
}

MarkupEvent::MarkupEvent(Type type)
: LocatedEvent(type, Location())
{
}

MarkupEvent::MarkupEvent(Type type, const Location &loc, Markup *markup)
: LocatedEvent(type, loc)
{
  if (markup)
    markup->moveTo(markup_);
}

StartElementEvent::StartElementEvent(const ElementType *elementType,
				     const ConstResourcePointer<Dtd> &dtd,
				     AttributeList *attributes,
				     const Location &startLocation,
				     Markup *markup)
: LocatedEvent(startElement, startLocation),
  elementType_(elementType),
  dtd_(dtd),
  included_(0),
  copied_(0),
  markup_(markup),
  attributes_(attributes)
{
}

StartElementEvent::~StartElementEvent()
{
  if (copied_) {
    delete attributes_;
    delete markup_;
  }
}

void StartElementEvent::copyData()
{
  if (!copied_) {
    {
      AttributeList *p = new AttributeList;
      attributes_->moveTo(*p);
      attributes_ = p;
    }
    {
      Markup *p = new Markup;
      markup_->moveTo(*p);
      markup_ = p;
    }
    copied_ = 1;
  }
}

EndElementEvent::EndElementEvent(const ElementType *elementType,
				 const ConstResourcePointer<Dtd> &dtd,
				 const Location &startLocation,
				 Markup *markup)
: LocatedEvent(endElement, startLocation),
  elementType_(elementType),
  dtd_(dtd),
  included_(0),
  copied_(0),
  markup_(markup)
{
}

EndElementEvent::~EndElementEvent()
{
  if (copied_)
    delete markup_;
}

void EndElementEvent::copyData()
{
  if (!copied_) {
    Markup *p = new Markup;
    markup_->moveTo(*p);
    markup_ = p;
    copied_ = 1;
  }
}

DataEvent::DataEvent(Type type, const Char *p, size_t length,
		     const Location &location)
: p_(p),length_(length), LocatedEvent(type, location)
{
}

const Entity *DataEvent::entity() const
{
  return 0;
}

Boolean DataEvent::isRe(unsigned long &) const
{
  return 0;
}

ImmediateDataEvent::ImmediateDataEvent(Type type, const Char *p, size_t length,
				       const Location &location,
				       Boolean copy)
: DataEvent(type, p, length, location), alloc_(0)
{
  if (copy) 
    ImmediateDataEvent::copyData();
}

ImmediateDataEvent::~ImmediateDataEvent()
{
  if (alloc_)
    delete [] alloc_;
}

void ImmediateDataEvent::copyData()
{
  if (!alloc_) {
    alloc_ = new Char[length_];
    memcpy(alloc_, p_, length_*sizeof(Char));
    p_ = alloc_;
  }
}

ReEvent::ReEvent(const Char *p, const Location &location, unsigned long serial)
: ImmediateDataEvent(characterData, p, 1, location, 0),
  serial_(serial)
{
}

Boolean ReEvent::isRe(unsigned long &serial) const
{
  serial = serial_;
  return 1;
}

DataEntityEvent::DataEntityEvent(Type type, const InternalDataEntity *entity,
				 const ConstResourcePointer<Origin> &origin)
: DataEvent(type,
	    entity->string().pointer(),
	    entity->string().length(),
	    Location(origin, 0))
{
}

const Entity *DataEntityEvent::entity() const
{
  return location().origin()->asEntityOrigin()->entity().pointer();
}

CdataEntityEvent::CdataEntityEvent(const InternalCdataEntity *entity,
				   const ConstResourcePointer<Origin> &origin)
: DataEntityEvent(characterData, entity, origin)
{
}

SdataEntityEvent::SdataEntityEvent(const InternalSdataEntity *entity,
				   const ConstResourcePointer<Origin> &origin)
: DataEntityEvent(sdataEntity, entity, origin)
{
}

MessageEvent::MessageEvent(Message &m)
: Event(Event::message)
{
  m.moveTo(message_);
}

PiEvent::PiEvent(const Char *data, size_t dataLength, const Location &location)
: data_(data), dataLength_(dataLength), LocatedEvent(pi, location)
{
}

const Entity *PiEvent::entity() const
{
  return 0;
}

PiEntityEvent::PiEntityEvent(const PiEntity *entity,
			     const ConstResourcePointer<Origin> &origin)
: PiEvent(entity->string().pointer(), entity->string().length(),
	  Location(origin, 0))
{
}

const Entity *PiEntityEvent::entity() const
{
  return location().origin()->asEntityOrigin()->entity().pointer();
}

ImmediatePiEvent::ImmediatePiEvent(CString &str, const Location &loc)
: PiEvent(str.pointer(), str.length(), loc)
{
  str.moveTo(string_);
}

ExternalEntityEvent::ExternalEntityEvent(Type type, 
					 const ConstResourcePointer<EntityOrigin> &origin)
: origin_(origin), Event(type)
{
}

ExternalDataEntityEvent::ExternalDataEntityEvent(const ExternalDataEntity *entity,
						 const ConstResourcePointer<EntityOrigin> &origin)
: dataEntity_(entity), ExternalEntityEvent(externalDataEntity, origin)
{
}

SubdocEntityEvent::SubdocEntityEvent(const SubdocEntity *entity,
				     const ConstResourcePointer<EntityOrigin> &origin)
: subdocEntity_(entity), ExternalEntityEvent(subdocEntity, origin)
{
}

AppinfoEvent::AppinfoEvent(const Location &location)
: LocatedEvent(appinfo, location), appinfoNone_(1)
{
}

AppinfoEvent::AppinfoEvent(const Text &text, const Location &location)
: LocatedEvent(appinfo, location), appinfoNone_(0), appinfo_(text)
{
}

Boolean AppinfoEvent::literal(const CString *&p) const
{
  if (appinfoNone_)
    return 0;
  p = &appinfo_.string();
  return 1;
}

SimpleLinkEvent::SimpleLinkEvent(const CString &name,
				 AttributeList &attributes,
				 const Location &location)
: LocatedEvent(simpleLink, location), name_(name)
{
  attributes.moveTo(attributes_);
}

ComplexLinkEvent::ComplexLinkEvent(const ConstResourcePointer<ComplexLpd> &lpd)
: Event(complexLink), lpd_(lpd)
{
}

UselinkEvent::UselinkEvent(const ConstResourcePointer<Lpd> &lpd,
			   const LinkSet *linkSet,
			   Boolean restore,
			   const Location &loc,
			   Markup *markup)
: MarkupEvent(uselink, loc, markup),
  lpd_(lpd),
  linkSet_(linkSet),
  restore_(restore)
{
}

UsemapEvent::UsemapEvent(const ShortReferenceMap *map,
			 Vector<const ElementType *> &elements,
			 const ConstResourcePointer<Dtd> &dtd,
			 const Location &loc,
			 Markup *markup)
: MarkupEvent(usemap, loc, markup),
  map_(map),
  dtd_(dtd)
{
  elements.moveTo(elements_);
}

StartSubsetEvent::StartSubsetEvent(Type type,
				   const CString &name,
				   const ConstResourcePointer<Entity> &entity,
				   Boolean hasInternalSubset,
				   const Location &loc,
				   Markup *markup)
: name_(name), entity_(entity), hasInternalSubset_(hasInternalSubset),
  MarkupEvent(type, loc, markup)
{
}

StartDtdEvent::StartDtdEvent(const CString &name,
			     const ConstResourcePointer<Entity> &entity,
			     Boolean hasInternalSubset,
			     const Location &loc,
			     Markup *markup)
: StartSubsetEvent(startDtd, name, entity, hasInternalSubset, loc, markup)
{
}

StartLpdEvent::StartLpdEvent(const CString &name,
			     const ConstResourcePointer<Entity> &entity,
			     Boolean hasInternalSubset,
			     const Location &loc,
			     Markup *markup)
: StartSubsetEvent(startLpd, name, entity, hasInternalSubset, loc, markup)
{
}

EndDtdEvent::EndDtdEvent(const ConstResourcePointer<Dtd> &dtd,
			 const Location &loc,
			 Markup *markup)
:  MarkupEvent(endDtd, loc, markup), dtd_(dtd)
{
}

EndLpdEvent::EndLpdEvent(const ConstResourcePointer<Lpd> &lpd,
			 const Location &loc,
			 Markup *markup)
:  MarkupEvent(endLpd, loc, markup), lpd_(lpd)
{
}

EndPrologEvent::EndPrologEvent(const ConstResourcePointer<Dtd> &dtd,
			       const Location &location)
: LocatedEvent(endProlog, location), dtd_(dtd)
{
}

SgmlDeclEvent::SgmlDeclEvent(const ConstResourcePointer<Sd> &sd,
			     const ConstResourcePointer<Syntax> &syntax)

: sd_(sd), prologSyntax_(syntax), instanceSyntax_(syntax),
  nextIndex_(0), inDocumentEntity_(0),
  MarkupEvent(sgmlDecl)
{
}

SgmlDeclEvent::SgmlDeclEvent(const ConstResourcePointer<Sd> &sd,
			     const ConstResourcePointer<Syntax> &prologSyntax,
			     const ConstResourcePointer<Syntax> &instanceSyntax,
			     const ConstResourcePointer<Sd> &refSd,
			     const ConstResourcePointer<Syntax> &refSyntax,
			     Index nextIndex,
			     Boolean inDocumentEntity,
			     const Location &loc,
			     Markup *markup)
: sd_(sd), prologSyntax_(prologSyntax), instanceSyntax_(instanceSyntax),
  refSd_(refSd), refSyntax_(refSyntax),
  nextIndex_(nextIndex), inDocumentEntity_(inDocumentEntity),
  MarkupEvent(sgmlDecl, loc, markup)
{
}

CommentDeclEvent::CommentDeclEvent(const Location &loc,
				   Markup *markup)
: MarkupEvent(commentDecl, loc, markup)
{
}

SSepEvent::SSepEvent(const Char *p, size_t length,
		     const Location &location, Boolean copy)
: ImmediateDataEvent(sSep, p, length, location, copy)
{
}

IgnoredRsEvent::IgnoredRsEvent(Char c, const Location &location)
: LocatedEvent(ignoredRs, location), c_(c)
{
}

IgnoredReEvent::IgnoredReEvent(Char c, const Location &location,
			       unsigned long serial)
: LocatedEvent(ignoredRe, location),
  c_(c),
  serial_(serial)
{
}

ReOriginEvent::ReOriginEvent(Char c, const Location &location,
			     unsigned long serial)
: LocatedEvent(reOrigin, location), c_(c), serial_(serial)
{
}


IgnoredCharsEvent::IgnoredCharsEvent(const Char *p, size_t length,
				     const Location &location, Boolean copy)
: ImmediateDataEvent(ignoredChars, p, length, location, copy)
{
}

MarkedSectionEvent::MarkedSectionEvent(Type type, Status status,
				       const Location &loc,
				       Markup *markup)
: MarkupEvent(type, loc, markup),
  status_(status)
{
}

MarkedSectionStartEvent::MarkedSectionStartEvent(Status status,
						 const Location &loc,
						 Markup *markup)
: MarkedSectionEvent(markedSectionStart, status, loc, markup)
{
}

MarkedSectionEndEvent::MarkedSectionEndEvent(Status status,
					     const Location &loc,
					     Markup *markup)
: MarkedSectionEvent(markedSectionEnd, status, loc, markup)
{
}

EntityStartEvent::EntityStartEvent(const ConstResourcePointer<EntityOrigin> &origin)
: Event(entityStart), origin_(origin)
{
}

EntityEndEvent::EntityEndEvent(const Location &location)
: LocatedEvent(entityEnd, location)
{
}

EntityDeclEvent:: EntityDeclEvent(const ConstResourcePointer<Entity> &entity,
				  Boolean duplicate, const Location &loc,
				  Markup *markup)
: MarkupEvent(entityDecl, loc, markup),
  entity_(entity),
  duplicate_(duplicate)
{
}

NotationDeclEvent:: NotationDeclEvent(const ConstResourcePointer<Notation> &notation,
				      const Location &loc,
				      Markup *markup)
: MarkupEvent(notationDecl, loc, markup), notation_(notation)
{
}

ElementDeclEvent::ElementDeclEvent(Vector<const ElementType *> &elements,
				   const ConstResourcePointer<Dtd> &dtd,
				   const Location &loc,
				   Markup *markup)
: MarkupEvent(elementDecl, loc, markup), dtd_(dtd)
{
  elements.moveTo(elements_);
}

AttlistDeclEvent::AttlistDeclEvent(Vector<const ElementType *> &elements,
				   const ConstResourcePointer<Dtd> &dtd,
				   const Location &loc,
				   Markup *markup)
: MarkupEvent(attlistDecl, loc, markup), dtd_(dtd)
{
  elements.moveTo(elements_);
}

AttlistNotationDeclEvent::AttlistNotationDeclEvent(
  Vector<ConstResourcePointer<Notation> > &notations, const Location &loc,
						   Markup *markup)
: MarkupEvent(attlistNotationDecl, loc, markup)
{
  notations.moveTo(notations_);
}

LinkAttlistDeclEvent::LinkAttlistDeclEvent(Vector<const ElementType *> &elements,
					   const ConstResourcePointer<ComplexLpd> &lpd,
					   const Location &loc,
					   Markup *markup)
: MarkupEvent(linkAttlistDecl, loc, markup), lpd_(lpd)
{
  elements.moveTo(elements_);
}

LinkDeclEvent::LinkDeclEvent(const LinkSet *linkSet,
			     const ConstResourcePointer<ComplexLpd> &lpd,
			     const Location &loc,
			     Markup *markup)
: MarkupEvent(linkDecl, loc, markup), lpd_(lpd), linkSet_(linkSet)
{
}

IdLinkDeclEvent::IdLinkDeclEvent(const ConstResourcePointer<ComplexLpd> &lpd,
				 const Location &loc,
				 Markup *markup)
: MarkupEvent(linkDecl, loc, markup), lpd_(lpd)
{
}

ShortrefDeclEvent::ShortrefDeclEvent(const ShortReferenceMap *map,
				     const ConstResourcePointer<Dtd> &dtd,
				     const Location &loc,
				     Markup *markup)
: MarkupEvent(shortrefDecl, loc, markup), map_(map), dtd_(dtd)
{
}

IgnoredMarkupEvent::IgnoredMarkupEvent(const Location &loc,
				       Markup *markup)
: MarkupEvent(ignoredMarkup, loc, markup)
{
}

EntityDefaultedEvent::EntityDefaultedEvent(const ConstResourcePointer<Entity> &entity,
					   const Location &loc)
: LocatedEvent(entityDefaulted, loc), entity_(entity)
{
}

EventHandler::~EventHandler()
{
}

EventQueue::EventQueue()
{
}

#define EVENT(c, f) \
  void EventHandler::f(c *event) { delete event; } \
  void EventQueue::f(c *event) { append(event); } \
  void c::handle(EventHandler &handler) { handler.f(this); }
#include "events.H"
#undef EVENT

Pass1EventHandler::Pass1EventHandler()
: hadError_(0), origHandler_(0)
{
}

void Pass1EventHandler::init(EventHandler *origHandler)
{
  hadError_ = 0;
  origHandler_ = origHandler;
}

void Pass1EventHandler::message(MessageEvent *event)
{
  if (event->message().isError()) {
    hadError_ = 1;
    origHandler_->message(event);
  }
  else
    IQueue<Event>::append(event);
}
