// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifndef Event_INCLUDED
#define Event_INCLUDED 1
#ifdef __GNUG__
#pragma interface
#endif

#include "Link.H"
#include "Allocator.H"
#include "Location.H"
#include "Vector.H"
#include "Owner.H"
#include "Boolean.H"
#include "types.H"
#include "ResourcePointer.H"
#include "CString.H"
#include "Notation.H"
#include "Sd.H"
#include "Syntax.H"
#include "Dtd.H"
#include "ElementType.H"
#include "Text.H"
#include "Lpd.H"
#include "Message.H"
#include "Markup.H"
#include "ShortReferenceMap.H"

class EventHandler;

class Event : public Link {
public:
  enum Type {
    message,
    characterData,
    startElement,
    endElement,
    pi,
    sdataEntity,
    externalDataEntity,
    subdocEntity,
    appinfo,
    startDtd,
    endDtd,
    startLpd,
    endLpd,
    endProlog,
    sgmlDecl,
    simpleLink,
    complexLink,
    uselink,
    usemap,
    commentDecl,
    sSep,
    ignoredRs,
    ignoredRe,
    reOrigin,
    ignoredChars,
    markedSectionStart,
    markedSectionEnd,
    entityStart,
    entityEnd,
    notationDecl,
    entityDecl,
    elementDecl,
    attlistDecl,		// not #NOTATION and not in LPD
    attlistNotationDecl,
    linkAttlistDecl,
    linkDecl,
    idLinkDecl,
    shortrefDecl,
    ignoredMarkup,
    entityDefaulted
    };
  Event(Type);
  virtual ~Event();
  virtual void handle(EventHandler &) = 0;
  virtual void copyData();
  void *operator new(size_t sz, Allocator &alloc) { return alloc.alloc(sz); }
  void *operator new(size_t sz) { return Allocator::allocSimple(sz); }
  void operator delete(void *p) { Allocator::free(p); }
  Type type() const;
private:
  Event(const Event &);		// undefined
  void operator=(const Event &); // undefined
  Type type_;
};

class LocatedEvent : public Event {
public:
  LocatedEvent(Type type, const Location &);
  const Location &location() const;
private:
  LocatedEvent(const LocatedEvent &); // undefined
  void operator=(const LocatedEvent &);	// undefined
  Location location_;
};

class MarkupEvent : public LocatedEvent {
public:
  MarkupEvent(Type type);
  MarkupEvent(Type type, const Location &, Markup *);
  const Markup &markup() const;
private:
  MarkupEvent(const MarkupEvent &); // undefined
  void operator=(const MarkupEvent &);	// undefined
  Markup markup_;
};

class MessageEvent : public Event {
public:
  MessageEvent(Message &);
  const Message &message() const;
  void handle(EventHandler &);
private:
  MessageEvent(const MessageEvent &); // undefined
  void operator=(const MessageEvent &);	// undefined
  Message message_;
};

class AttributeList;

class StartElementEvent : public LocatedEvent {
public:
  StartElementEvent(const ElementType *,
		    const ConstResourcePointer<Dtd> &,
		    AttributeList *,
		    const Location &,
		    Markup *);
  ~StartElementEvent();
  void handle(EventHandler &);
  Boolean mustOmitEnd() const;
  void setIncluded();
  Boolean included() const;
  const CString &name() const;
  const ElementType *elementType() const;
  const Markup *markupPtr() const;
  const AttributeList &attributes() const;
  void copyData();
private:
  StartElementEvent(const StartElementEvent &);	// undefined
  void operator=(const StartElementEvent &);	// undefined
  const ElementType *elementType_;
  ConstResourcePointer<Dtd> dtd_;
  PackedBoolean included_;
  PackedBoolean copied_;	// has copyData() been called
  Markup *markup_;
  AttributeList *attributes_;
};

class EndElementEvent : public LocatedEvent {
public:
  EndElementEvent(const ElementType *,
		  const ConstResourcePointer<Dtd> &,
		  const Location &,
		  Markup *);
  ~EndElementEvent();
  void handle(EventHandler &);
  void setIncluded();
  Boolean included() const;
  const CString &name() const;
  const ElementType *elementType() const;
  const Markup *markupPtr() const;
  void copyData();
private:
  EndElementEvent(const EndElementEvent &); // undefined
  void operator=(const EndElementEvent &);  // undefined
  const ElementType *elementType_;
  ConstResourcePointer<Dtd> dtd_;
  PackedBoolean included_;
  PackedBoolean copied_;	// has copyData() been called
  Markup *markup_;
};

class DataEvent : public LocatedEvent {
public:
  DataEvent(Type, const Char *, size_t, const Location &);
  void handle(EventHandler &);
  const Char *data() const;
  size_t dataLength() const;
  virtual Boolean isRe(unsigned long &serial) const;
  virtual const Entity *entity() const;
protected:
  const Char *p_;
  size_t length_;
private:
  DataEvent(const DataEvent &);	// undefined
  void operator=(const DataEvent &); // undefined
};

class ImmediateDataEvent : public DataEvent {
public:
  ImmediateDataEvent(Type type, const Char *, size_t, const Location &,
		     Boolean copy);
  ~ImmediateDataEvent();
  void copyData();
private:
  ImmediateDataEvent(const ImmediateDataEvent &); // undefined
  void operator=(const ImmediateDataEvent &);	  // undefined
  Char *alloc_;
};

class InternalDataEntity;

class DataEntityEvent : public DataEvent {
public:
  DataEntityEvent(Type type, const InternalDataEntity *,
		  const ConstResourcePointer<Origin> &);
  const Entity *entity() const;
private:
  DataEntityEvent(const DataEntityEvent &); // undefined
  void operator=(const DataEntityEvent &);  // undefined
};

class InternalCdataEntity;

class CdataEntityEvent : public DataEntityEvent {
public:
  CdataEntityEvent(const InternalCdataEntity *,
		   const ConstResourcePointer<Origin> &);
private:
  CdataEntityEvent(const CdataEntityEvent &); // undefined
  void operator=(const CdataEntityEvent &);   // undefined
};

class InternalSdataEntity;

class SdataEntityEvent : public DataEntityEvent {
public:
  SdataEntityEvent(const InternalSdataEntity *,
		   const ConstResourcePointer<Origin> &);
  void handle(EventHandler &);
private:
  SdataEntityEvent(const SdataEntityEvent &); // undefined
  void operator=(const SdataEntityEvent &);   // undefined
};

class PiEntity;

class PiEvent : public LocatedEvent {
public:
  PiEvent(const Char *, size_t, const Location &);
  const Char *data() const;
  size_t dataLength() const;
  virtual const Entity *entity() const;
  void handle(EventHandler &);
private:
  PiEvent(const PiEvent &);	// undefined
  void operator=(const PiEvent &); // undefined
  const Char *data_;
  size_t dataLength_;
};

class ImmediatePiEvent : public PiEvent {
public:
  ImmediatePiEvent(CString &, const Location &);
private:
  ImmediatePiEvent(const ImmediatePiEvent &); // undefined
  void operator=(const ImmediatePiEvent &);   // undefined
  CString string_;
};

class PiEntityEvent : public PiEvent {
public:
  PiEntityEvent(const PiEntity *entity,
		const ConstResourcePointer<Origin> &origin);
  const Entity *entity() const;
private:
  PiEntityEvent(const PiEntityEvent &);	// undefined
  void operator=(const PiEntityEvent &); // undefined
};

class ExternalNonTextEntity;
class ExternalDataEntity;
class SubdocEntity;

class ExternalEntityEvent : public Event {
public:
  ExternalEntityEvent(Type type,
		      const ConstResourcePointer<EntityOrigin> &);
  const ConstResourcePointer<EntityOrigin> &entityOrigin() const;
  const Location &location() const;
private:
  ExternalEntityEvent(const ExternalEntityEvent &); // undefined
  void operator=(const ExternalEntityEvent &);	    // undefined
  ConstResourcePointer<EntityOrigin> origin_;
};

class ExternalDataEntityEvent : public ExternalEntityEvent {
public:
  ExternalDataEntityEvent(const ExternalDataEntity *,
			  const ConstResourcePointer<EntityOrigin> &);
  void handle(EventHandler &);
  const ExternalDataEntity *entity() const;
private:
  ExternalDataEntityEvent(const ExternalDataEntityEvent &); // undefined
  void operator=(const ExternalDataEntityEvent &);	    // undefined
  const ExternalDataEntity *dataEntity_;
};

class SubdocEntityEvent : public ExternalEntityEvent {
public:
  SubdocEntityEvent(const SubdocEntity *,
		    const ConstResourcePointer<EntityOrigin> &);
  void handle(EventHandler &);
  const SubdocEntity *entity() const;
private:
  SubdocEntityEvent(const SubdocEntityEvent &);	// undefined
  void operator=(const SubdocEntityEvent &);	// undefined
  const SubdocEntity *subdocEntity_;
};

class AppinfoEvent : public LocatedEvent {
public:
  AppinfoEvent(const Location &);
  AppinfoEvent(const Text &, const Location &);
  void handle(EventHandler &);
  Boolean literal(const CString *&) const;
private:
  AppinfoEvent(const AppinfoEvent &); // undefined
  void operator=(const AppinfoEvent &);	// undefined
  Boolean appinfoNone_;
  Text appinfo_;
};

class SimpleLinkEvent : public LocatedEvent {
public:
  SimpleLinkEvent(const CString &,
		  AttributeList &,
		  const Location &);
  void handle(EventHandler &);
  const AttributeList &attributes() const;
  const CString &name() const;
private:
  SimpleLinkEvent(const SimpleLinkEvent &); // undefined
  void operator=(const SimpleLinkEvent &);
  CString name_;
  AttributeList attributes_;
};

class ComplexLinkEvent : public Event {
public:
  ComplexLinkEvent(const ConstResourcePointer<ComplexLpd> &);
  void handle(EventHandler &);
  const CString &name() const;
  const ConstResourcePointer<ComplexLpd> &lpd() const;
private:
  ComplexLinkEvent(const ComplexLinkEvent &); // undefined
  void operator=(const ComplexLinkEvent &);
  ConstResourcePointer<ComplexLpd> lpd_;
};

class UselinkEvent : public MarkupEvent {
public:
  UselinkEvent(const ConstResourcePointer<Lpd> &,
	       const LinkSet *,
	       Boolean restore,
	       const Location &,
	       Markup *);
  void handle(EventHandler &);
  const ConstResourcePointer<Lpd> &lpd() const;
  const LinkSet *linkSet() const;
  Boolean restore() const;
private:
  UselinkEvent(const UselinkEvent &); // undefined
  void operator=(const UselinkEvent &);	// undefined
  ConstResourcePointer<Lpd> lpd_;
  const LinkSet *linkSet_;
  Boolean restore_;
};

class UsemapEvent : public MarkupEvent {
public:
  UsemapEvent(const ShortReferenceMap *,
	      Vector<const ElementType *> &,
	      const ConstResourcePointer<Dtd> &,
	      const Location &,
	      Markup *);
  void handle(EventHandler &);
  const ShortReferenceMap *map() const;
  const Vector<const ElementType *> &elements() const;
private:
  UsemapEvent(const UsemapEvent &); // undefined
  void operator=(const UsemapEvent &); // undefined
  ConstResourcePointer<Dtd> dtd_;
  Vector<const ElementType *> elements_;
  const ShortReferenceMap *map_;
};

class StartSubsetEvent : public MarkupEvent {
public:
  StartSubsetEvent(Type,
		   const CString &,
		   const ConstResourcePointer<Entity> &entity,
		   Boolean hasInternalSubset,
		   const Location &,
		   Markup *);
  const CString &name() const;
  const ConstResourcePointer<Entity> &entity() const;
  Boolean hasInternalSubset() const;
private:
  StartSubsetEvent(const StartSubsetEvent &);	// undefined
  void operator=(const StartSubsetEvent &); // undefined
  CString name_;
  ConstResourcePointer<Entity> entity_;
  Boolean hasInternalSubset_;
};

class StartDtdEvent : public StartSubsetEvent {
public:
  StartDtdEvent(const CString &,
		const ConstResourcePointer<Entity> &entity,
		Boolean hasInternalSubset,
		const Location &,
		Markup *);
  void handle(EventHandler &);
private:
  StartDtdEvent(const StartDtdEvent &);	// undefined
  void operator=(const StartDtdEvent &); // undefined
};

class StartLpdEvent : public StartSubsetEvent {
public:
  StartLpdEvent(const CString &,
		const ConstResourcePointer<Entity> &entity,
		Boolean hasInternalSubset,
		const Location &,
		Markup *);
  void handle(EventHandler &);
private:
  StartLpdEvent(const StartLpdEvent &);	// undefined
  void operator=(const StartLpdEvent &); // undefined
};

class EndDtdEvent : public MarkupEvent {
public:
  EndDtdEvent(const ConstResourcePointer<Dtd> &, const Location &,
	      Markup *);
  void handle(EventHandler &);
  const Dtd &dtd() const;
  const ConstResourcePointer<Dtd> &dtdPointer() const;
private:
  EndDtdEvent(const EndDtdEvent &); // undefined
  void operator=(const EndDtdEvent &); // undefined
  ConstResourcePointer<Dtd> dtd_;
};

class EndLpdEvent : public MarkupEvent {
public:
  EndLpdEvent(const ConstResourcePointer<Lpd> &, const Location &,
	      Markup *);
  void handle(EventHandler &);
  const Lpd &lpd() const;
  const ConstResourcePointer<Lpd> &lpdPointer() const;
private:
  EndLpdEvent(const EndLpdEvent &); // undefined
  void operator=(const EndLpdEvent &); // undefined
  ConstResourcePointer<Lpd> lpd_;
};

class EndPrologEvent : public LocatedEvent {
public:
  EndPrologEvent(const ConstResourcePointer<Dtd> &dtd, const Location &);
  void handle(EventHandler &);
  const Dtd &dtd() const;
  const ConstResourcePointer<Dtd> &dtdPointer() const;
private:
  EndPrologEvent(const EndPrologEvent &); // undefined
  void operator=(const EndPrologEvent &); // undefined
  ConstResourcePointer<Dtd> dtd_;
};

class SgmlDeclEvent : public MarkupEvent {
public:
  // for an implied SGML declaration
  SgmlDeclEvent(const ConstResourcePointer<Sd> &,
		const ConstResourcePointer<Syntax> &syntax);
  // for an explicit SGML declaration
  SgmlDeclEvent(const ConstResourcePointer<Sd> &,
		const ConstResourcePointer<Syntax> &syntax,
		const ConstResourcePointer<Syntax> &instanceSyntax,
		const ConstResourcePointer<Sd> &refSd,
		const ConstResourcePointer<Syntax> &refSyntax,
		Index nextIndex,
		Boolean inDocumentEntity,
		const Location &,
		Markup *);
  void handle(EventHandler &);
  const Sd &sd() const;
  const ConstResourcePointer<Sd> &sdPointer() const;
  const Syntax &prologSyntax() const;
  const ConstResourcePointer<Syntax> &prologSyntaxPointer() const;
  const Syntax &instanceSyntax() const;
  const ConstResourcePointer<Syntax> &instanceSyntaxPointer() const;
  const ConstResourcePointer<Sd> &refSdPointer() const;
  const ConstResourcePointer<Syntax> &refSyntaxPointer() const;
  Boolean inDocumentEntity() const;
private:
  SgmlDeclEvent(const SgmlDeclEvent &);	// undefined
  void operator=(const SgmlDeclEvent &); // undefined
  ConstResourcePointer<Sd> sd_;
  ConstResourcePointer<Syntax> prologSyntax_;
  ConstResourcePointer<Syntax> instanceSyntax_;
  ConstResourcePointer<Sd> refSd_;
  ConstResourcePointer<Syntax> refSyntax_;
  Index nextIndex_;
  Boolean inDocumentEntity_;
};

class CommentDeclEvent : public MarkupEvent {
public:
  CommentDeclEvent(const Location &, Markup *);
  void handle(EventHandler &);
private:
  CommentDeclEvent(const CommentDeclEvent &); // undefined
  void operator=(const CommentDeclEvent &);   // undefined
};

class SSepEvent : public ImmediateDataEvent {
public:
  SSepEvent(const Char *, size_t, const Location &, Boolean copy);
  void handle(EventHandler &);
private:
  SSepEvent(const SSepEvent &);	// undefined
  void operator=(const SSepEvent &); // undefined
};

class IgnoredRsEvent : public LocatedEvent {
public:
  IgnoredRsEvent(Char c, const Location &);
  void handle(EventHandler &);
  Char rs() const;
private:
  IgnoredRsEvent(const IgnoredRsEvent &); // undefined
  void operator=(const IgnoredRsEvent &); // undefined
  Char c_;
};

class IgnoredReEvent : public LocatedEvent {
public:
  IgnoredReEvent(Char c, const Location &, unsigned long serial);
  void handle(EventHandler &);
  Char re() const;
  unsigned long serial() const;
private:
  IgnoredReEvent(const IgnoredReEvent &); // undefined
  void operator=(const IgnoredReEvent &); // undefined
  unsigned long serial_;
  Char c_;
};

class ReEvent : public ImmediateDataEvent {
public:
  ReEvent(const Char *, const Location &, unsigned long serial);
  Boolean isRe(unsigned long &serial) const;
private:
  ReEvent(const ReEvent &);	// undefined
  void operator=(const ReEvent &); // undefined
  unsigned long serial_;
};

class ReOriginEvent : public LocatedEvent {
public:
  ReOriginEvent(Char c, const Location &, unsigned long serial);
  void handle(EventHandler &);
  Char re() const;
  unsigned long serial() const;
private:
  ReOriginEvent(const ReOriginEvent &);	// undefined
  void operator=(const ReOriginEvent &); // undefined
  unsigned long serial_;
  Char c_;
};

class IgnoredCharsEvent : public ImmediateDataEvent {
public:
  IgnoredCharsEvent(const Char *, size_t, const Location &, Boolean copy);
  void handle(EventHandler &);
private:
  IgnoredCharsEvent(const IgnoredCharsEvent &);	// undefined
  void operator=(const IgnoredCharsEvent &);	// undefined
};

class MarkedSectionEvent : public MarkupEvent {
public:
  enum Status { include, rcdata, cdata, ignore }; // in priority order
  MarkedSectionEvent(Type, Status, const Location &, Markup *);
  Status status() const;
private:
  MarkedSectionEvent(const MarkedSectionEvent &); // undefined
  void operator=(const MarkedSectionEvent &);	  // undefined
  Status status_;
};

class MarkedSectionStartEvent : public MarkedSectionEvent {
public:
  MarkedSectionStartEvent(Status, const Location &, Markup *);
  void handle(EventHandler &);
private:
  MarkedSectionStartEvent(const MarkedSectionStartEvent &); // undefined
  void operator=(const MarkedSectionStartEvent &);	    // undefined
};

class MarkedSectionEndEvent : public MarkedSectionEvent {
public:
  MarkedSectionEndEvent(Status, const Location &, Markup *);
  void handle(EventHandler &);
private:
  MarkedSectionEndEvent(const MarkedSectionEndEvent &); // undefined
  void operator=(const MarkedSectionEndEvent &);	    // undefined
};

class EntityStartEvent : public Event {
public:
  EntityStartEvent(const ConstResourcePointer<EntityOrigin> &origin);
  void handle(EventHandler &);
  const Entity *entity() const;
  const ConstResourcePointer<EntityOrigin> &entityOrigin() const;
private:
  EntityStartEvent(const EntityStartEvent &); // undefined
  void operator=(const EntityStartEvent &); // undefined

  ConstResourcePointer<EntityOrigin> origin_;
};

class EntityEndEvent : public LocatedEvent {
public:
  EntityEndEvent(const Location &);
  void handle(EventHandler &);
private:
  EntityEndEvent(const EntityEndEvent &); // undefined
  void operator=(const EntityEndEvent &); // undefined
};

class EntityDeclEvent : public MarkupEvent {
public:
  EntityDeclEvent(const ConstResourcePointer<Entity> &,
		  Boolean duplicate,
		  const Location &,
		  Markup *);
  void handle(EventHandler &);
  const Entity &entity() const;
  const ConstResourcePointer<Entity> &entityPointer() const;
  Boolean duplicate() const;
  // The name of the entity will be empty if this is the default entity.
private:
  Boolean duplicate_;
  // This will actually point to an external entity.
  ConstResourcePointer<Entity> entity_;
};

class NotationDeclEvent : public MarkupEvent {
public:
  NotationDeclEvent(const ConstResourcePointer<Notation> &,
		    const Location &,
		    Markup *);
  void handle(EventHandler &);
  const Notation &notation() const;
  const ConstResourcePointer<Notation> &notationPointer() const;
private:
  NotationDeclEvent(const NotationDeclEvent &);	// undefined
  void operator=(const NotationDeclEvent &);	// undefined
  ConstResourcePointer<Notation> notation_;
};

class ElementDeclEvent : public MarkupEvent {
public:
  ElementDeclEvent(Vector<const ElementType *> &elements,
		   const ConstResourcePointer<Dtd> &,
		   const Location &,
		   Markup *);
  void handle(EventHandler &);
  const Vector<const ElementType *> &elements() const;
private:
  ElementDeclEvent(const ElementDeclEvent &); // undefined
  void operator=(const ElementDeclEvent &);   // undefined
  Vector<const ElementType *> elements_;
  ConstResourcePointer<Dtd> dtd_;
};

class AttlistDeclEvent : public MarkupEvent {
public:
  AttlistDeclEvent(Vector<const ElementType *> &elements,
		   const ConstResourcePointer<Dtd> &,
		   const Location &,
		   Markup *);
  void handle(EventHandler &);
  const Vector<const ElementType *> &elements() const;
private:
  AttlistDeclEvent(const AttlistDeclEvent &); // undefined
  void operator=(const AttlistDeclEvent &);   // undefined
  Vector<const ElementType *> elements_;
  ConstResourcePointer<Dtd> dtd_;
};

class AttlistNotationDeclEvent : public MarkupEvent {
public:
  AttlistNotationDeclEvent(Vector<ConstResourcePointer<Notation> > &notations,
			   const Location &,
			   Markup *);
  void handle(EventHandler &);
  const Vector<ConstResourcePointer<Notation> > &notations() const;
private:
  AttlistNotationDeclEvent(const AttlistNotationDeclEvent &); // undefined
  void operator=(const AttlistDeclEvent &);		      // undefined
  Vector<ConstResourcePointer<Notation> > notations_;
};

class LinkAttlistDeclEvent : public MarkupEvent {
public:
  LinkAttlistDeclEvent(Vector<const ElementType *> &elements,
		       const ConstResourcePointer<ComplexLpd> &,
		       const Location &,
		       Markup *);
  void handle(EventHandler &);
  const Vector<const ElementType *> &elements() const;
  const ComplexLpd &lpd() const;
private:
  LinkAttlistDeclEvent(const LinkAttlistDeclEvent &); // undefined
  void operator=(const LinkAttlistDeclEvent &);   // undefined
  Vector<const ElementType *> elements_;
  ConstResourcePointer<ComplexLpd> lpd_;
};

class LinkDeclEvent : public MarkupEvent {
public:
  LinkDeclEvent(const LinkSet *linkSet,
		const ConstResourcePointer<ComplexLpd> &,
		const Location &,
		Markup *);
  void handle(EventHandler &);
  const LinkSet *linkSet() const;
  const ComplexLpd &lpd() const;
private:
  LinkDeclEvent(const LinkDeclEvent &); // undefined
  void operator=(const LinkDeclEvent &); // undefined
  const LinkSet *linkSet_;
  ConstResourcePointer<ComplexLpd> lpd_;
};

class IdLinkDeclEvent : public MarkupEvent {
public:
  IdLinkDeclEvent(const ConstResourcePointer<ComplexLpd> &,
		  const Location &,
		  Markup *);
  void handle(EventHandler &);
  const ComplexLpd &lpd() const;
private:
  IdLinkDeclEvent(const IdLinkDeclEvent &); // undefined
  void operator=(const IdLinkDeclEvent &); // undefined
  ConstResourcePointer<ComplexLpd> lpd_;
};

class ShortrefDeclEvent : public MarkupEvent {
public:
  ShortrefDeclEvent(const ShortReferenceMap *,
		    const ConstResourcePointer<Dtd> &,
		    const Location &,
		    Markup *);
  void handle(EventHandler &);
  const ShortReferenceMap *map() const;
private:
  ShortrefDeclEvent(const ShortrefDeclEvent &);	// undefined
  void operator=(const ShortrefDeclEvent &);	// undefined
  const ShortReferenceMap *map_;
  ConstResourcePointer<Dtd> dtd_;
};

class IgnoredMarkupEvent : public MarkupEvent {
public:
  IgnoredMarkupEvent(const Location &, Markup *);
  void handle(EventHandler &);
private:
  IgnoredMarkupEvent(const IgnoredMarkupEvent &); // undefined
  void operator=(const IgnoredMarkupEvent &);	  // undefined
};

// This is for an undeclared entity whose first occurrence
// is in the instance, when there is a default entity:
// ie it extends the namespace of general entities after
// the end of the prolog.

class EntityDefaultedEvent : public LocatedEvent {
public:
  EntityDefaultedEvent(const ConstResourcePointer<Entity> &,
		       const Location &);
  void handle(EventHandler &);
  const Entity &entity() const;
  const ConstResourcePointer<Entity> &entityPointer() const;
private:
  EntityDefaultedEvent(const EntityDefaultedEvent &); // undefined
  void operator=(const EntityDefaultedEvent &);	      // undefined
  ConstResourcePointer<Entity> entity_;
};

class EventHandler {
public:
  virtual ~EventHandler();
  virtual void message(MessageEvent *) = 0;
  virtual void data(DataEvent *);
  virtual void startElement(StartElementEvent *);
  virtual void endElement(EndElementEvent *);
  virtual void pi(PiEvent *);
  virtual void sdataEntity(SdataEntityEvent *);
  virtual void externalDataEntity(ExternalDataEntityEvent *);
  virtual void subdocEntity(SubdocEntityEvent *);
  virtual void appinfo(AppinfoEvent *);
  virtual void simpleLink(SimpleLinkEvent *);
  virtual void complexLink(ComplexLinkEvent *);
  virtual void uselink(UselinkEvent *);
  virtual void usemap(UsemapEvent *);
  virtual void startDtd(StartDtdEvent *);
  virtual void endDtd(EndDtdEvent *);
  virtual void startLpd(StartLpdEvent *);
  virtual void endLpd(EndLpdEvent *);
  virtual void endProlog(EndPrologEvent *);
  virtual void sgmlDecl(SgmlDeclEvent *);
  virtual void commentDecl(CommentDeclEvent *);
  virtual void sSep(SSepEvent *);
  virtual void ignoredRs(IgnoredRsEvent *);
  virtual void ignoredRe(IgnoredReEvent *);
  virtual void reOrigin(ReOriginEvent *);
  virtual void ignoredChars(IgnoredCharsEvent *);
  virtual void markedSectionStart(MarkedSectionStartEvent *);
  virtual void markedSectionEnd(MarkedSectionEndEvent *);
  virtual void entityStart(EntityStartEvent *);
  virtual void entityEnd(EntityEndEvent *);
  virtual void notationDecl(NotationDeclEvent *);
  virtual void entityDecl(EntityDeclEvent *);
  virtual void elementDecl(ElementDeclEvent *);
  virtual void attlistDecl(AttlistDeclEvent *);
  virtual void linkAttlistDecl(LinkAttlistDeclEvent *);
  virtual void attlistNotationDecl(AttlistNotationDeclEvent *);
  virtual void linkDecl(LinkDeclEvent *);
  virtual void idLinkDecl(IdLinkDeclEvent *);
  virtual void shortrefDecl(ShortrefDeclEvent *);
  virtual void ignoredMarkup(IgnoredMarkupEvent *);
  virtual void entityDefaulted(EntityDefaultedEvent *);
};

inline
Event::Event(Type type)
: type_(type)
{
}

inline
Event::Type Event::type() const
{
  return type_;
}

inline
const Location &LocatedEvent::location() const
{
  return location_;
}

inline
const Markup &MarkupEvent::markup() const
{
  return markup_;
}

inline
const Message &MessageEvent::message() const
{
  return message_;
}

inline
const ElementType *StartElementEvent::elementType() const
{
  return elementType_;
}

inline
const CString &StartElementEvent::name() const
{
  return elementType_->name();
}

inline
void StartElementEvent::setIncluded()
{
  included_ = 1;
}

inline
Boolean StartElementEvent::included() const
{
  return included_;
}

inline
const Markup *StartElementEvent::markupPtr() const
{
  return markup_;
}

inline
const AttributeList &StartElementEvent::attributes() const
{
  return *attributes_;
}

inline
Boolean StartElementEvent::mustOmitEnd() const
{
  return ((elementType()->definition()->declaredContent()
	   == ElementDefinition::empty)
	  ||  attributes_->conref());
}

inline
const ElementType *EndElementEvent::elementType() const
{
  return elementType_;
}

inline
const CString &EndElementEvent::name() const
{
  return elementType_->name();
}

inline
void EndElementEvent::setIncluded()
{
  included_ = 1;
}

inline
Boolean EndElementEvent::included() const
{
  return included_;
}

inline
const Markup *EndElementEvent::markupPtr() const
{
  return markup_;
}

inline
const Char *DataEvent::data() const
{
  return p_;
}

inline
size_t DataEvent::dataLength() const
{
  return length_;
}

inline
const Char *PiEvent::data() const
{
  return data_;
}

inline
size_t PiEvent::dataLength() const
{
  return dataLength_;
}

inline
const ConstResourcePointer<EntityOrigin> &
ExternalEntityEvent::entityOrigin() const
{
  return origin_;
}

inline
const Location &ExternalEntityEvent::location() const
{
  return origin_->parent();
}

inline
const ExternalDataEntity *ExternalDataEntityEvent::entity() const
{
  return dataEntity_;
}

inline
const SubdocEntity *SubdocEntityEvent::entity() const
{
  return subdocEntity_;
}

inline
const AttributeList &SimpleLinkEvent::attributes() const
{
  return attributes_;
}

inline
const CString &SimpleLinkEvent::name() const
{
  return name_;
}

inline
const CString &ComplexLinkEvent::name() const
{
  return lpd_->name();
}

inline
const ConstResourcePointer<ComplexLpd> &ComplexLinkEvent::lpd() const
{
  return lpd_;
}

inline
const ConstResourcePointer<Lpd> &UselinkEvent::lpd() const
{
  return lpd_;
}

inline
const LinkSet *UselinkEvent::linkSet() const
{
  return linkSet_;
}

inline
Boolean UselinkEvent::restore() const
{
  return restore_;
}

inline
const ShortReferenceMap *UsemapEvent::map() const
{
  return map_;
}

inline
const CString &StartSubsetEvent::name() const
{
  return name_;
}

inline
const ConstResourcePointer<Entity> &StartSubsetEvent::entity() const
{
  return entity_;
}

inline
Boolean StartSubsetEvent::hasInternalSubset() const
{
  return hasInternalSubset_;
}

inline
const Dtd &EndDtdEvent::dtd() const
{
  return *dtd_;
}

inline
const ConstResourcePointer<Dtd> &EndDtdEvent::dtdPointer() const
{
  return dtd_;
}

inline
const Lpd &EndLpdEvent::lpd() const
{
  return *lpd_;
}

inline
const ConstResourcePointer<Lpd> &EndLpdEvent::lpdPointer() const
{
  return lpd_;
}

inline
const Dtd &EndPrologEvent::dtd() const
{
  return *dtd_;
}

inline
const ConstResourcePointer<Dtd> &EndPrologEvent::dtdPointer() const
{
  return dtd_;
}

inline
const Sd &SgmlDeclEvent::sd() const
{
  return *sd_;
}

inline
const ConstResourcePointer<Sd> &SgmlDeclEvent::sdPointer() const
{
  return sd_;
}

inline
const ConstResourcePointer<Sd> &SgmlDeclEvent::refSdPointer() const
{
  return refSd_;
}

inline
const Syntax &SgmlDeclEvent::prologSyntax() const
{
  return *prologSyntax_;
}

inline
const ConstResourcePointer<Syntax> &SgmlDeclEvent::prologSyntaxPointer() const
{
  return prologSyntax_;
}

inline
const Syntax &SgmlDeclEvent::instanceSyntax() const
{
  return *instanceSyntax_;
}

inline
const ConstResourcePointer<Syntax> &SgmlDeclEvent::instanceSyntaxPointer() const
{
  return instanceSyntax_;
}

inline
const ConstResourcePointer<Syntax> &SgmlDeclEvent::refSyntaxPointer() const
{
  return refSyntax_;
}

inline
Boolean SgmlDeclEvent::inDocumentEntity() const
{
  return inDocumentEntity_;
}

inline
Char IgnoredRsEvent::rs() const
{
  return c_;
}

inline
Char IgnoredReEvent::re() const
{
  return c_;
}

inline
unsigned long IgnoredReEvent::serial() const
{
  return serial_;
}

inline
Char ReOriginEvent::re() const
{
  return c_;
}

inline
unsigned long ReOriginEvent::serial() const
{
  return serial_;
}

inline
MarkedSectionEvent::Status MarkedSectionEvent::status() const
{
  return status_;
}

inline
const Entity *EntityStartEvent::entity() const
{
  return origin_->entity().pointer();
}

inline
const ConstResourcePointer<EntityOrigin> &
EntityStartEvent::entityOrigin() const
{
  return origin_;
}

inline
const ConstResourcePointer<Entity> &EntityDeclEvent::entityPointer() const
{
  return entity_;
}

inline
const Entity &EntityDeclEvent::entity() const
{
  return *entity_;
}

inline
Boolean EntityDeclEvent::duplicate() const
{
  return duplicate_;
}

inline
const Notation &NotationDeclEvent::notation() const
{
  return *notation_;
}

inline
const ConstResourcePointer<Notation> &NotationDeclEvent::notationPointer() const
{
  return notation_;
}

inline
const Vector<const ElementType *> &ElementDeclEvent::elements() const
{
  return elements_;
}

inline
const Vector<const ElementType *> &AttlistDeclEvent::elements() const
{
  return elements_;
}

inline
const Vector<const ElementType *> &LinkAttlistDeclEvent::elements() const
{
  return elements_;
}

inline
const ComplexLpd &LinkAttlistDeclEvent::lpd() const
{
  return *lpd_;
}

inline
const LinkSet *LinkDeclEvent::linkSet() const
{
  return linkSet_;
}

inline
const ComplexLpd &LinkDeclEvent::lpd() const
{
  return *lpd_;
}

inline
const ComplexLpd &IdLinkDeclEvent::lpd() const
{
  return *lpd_;
}

inline
const Vector<ConstResourcePointer<Notation> > &
AttlistNotationDeclEvent::notations() const
{
  return notations_;
}

inline
const ShortReferenceMap *ShortrefDeclEvent::map() const
{
  return map_;
}

inline
const Entity &EntityDefaultedEvent::entity() const
{
  return *entity_;
}

inline
const ConstResourcePointer<Entity> &EntityDefaultedEvent::entityPointer()
     const
{
  return entity_;
}

#endif /* not Event_INCLUDED */
