// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "Markup.H"
#include "InputSource.H"
#include "Location.H"
#include "macros.H"
#include "Entity.H"

MarkupItem::MarkupItem()
: type(Markup::delimiter)
{
}

MarkupItem::~MarkupItem()
{
  switch (type) {
  case Markup::entityStart:
    delete origin;
    break;
  case Markup::literal:
    delete text;
    break;
  case Markup::sdLiteral:
    delete sdText;
    break;
  }
}

MarkupItem::MarkupItem(const MarkupItem &item)
: type(item.type), index(item.index)
{
  switch (item.type) {
  case Markup::entityStart:
    origin = new ConstResourcePointer<Origin>(*item.origin);
    break;
  case Markup::literal:
    text = new Text(*item.text);
    break;
  case Markup::sdLiteral:
    sdText = new SdText(*item.sdText);
    break;
  default:
    nChars = item.nChars;
    break;
  }
}

void MarkupItem::operator=(const MarkupItem &item)
{
  switch (type) {
  case Markup::entityStart:
    if (item.type == Markup::entityStart) {
      *origin = *item.origin;
      return;
    }
    delete origin;
    break;
  case Markup::literal:
    if (item.type == Markup::literal) {
      *text = *item.text;
      return;
    }
    delete text;
    break;
  case Markup::sdLiteral:
    if (item.type == Markup::sdLiteral) {
      *sdText = *item.sdText;
      return;
    }
    delete sdText;
    break;
  }
  type = item.type;
  index = item.index;
  switch (item.type) {
  case Markup::entityStart:
    origin = new ConstResourcePointer<Origin>(*item.origin);
    break;
  case Markup::literal:
    text = new Text(*item.text);
    break;
  case Markup::sdLiteral:
    sdText = new SdText(*item.sdText);
    break;
  default:
    nChars = item.nChars;
    break;
  }
}

Markup::Markup()
{
}

void Markup::setLength(size_t n)
{
  size_t chopChars = 0;
  for (size_t i = n; i < items_.length(); i++)
    switch (items_[i].type) {
    case Markup::reservedName:
    case Markup::sdReservedName:
    case Markup::name:
    case Markup::nameToken:
    case Markup::number:
    case Markup::attributeValue:
    case Markup::s:
    case Markup::comment:
    case Markup::shortref:
      chopChars += items_[i].nChars;
      break;
    }
  items_.setLength(n);
  chars_.setLength(chars_.length() - chopChars);
}

void Markup::addDelim(Syntax::DelimGeneral d)
{
  MarkupItem &item = items_.grow();
  item.type = Markup::delimiter;
  item.index = d;
}

void Markup::addReservedName(Syntax::ReservedName rn, const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::reservedName;
  item.index = rn;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addSdReservedName(Sd::ReservedName rn, const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::sdReservedName;
  item.index = rn;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addSdReservedName(Sd::ReservedName rn,
			       const Char *str, size_t length)
{
  MarkupItem &item = items_.grow();
  item.nChars = length;
  item.type = Markup::sdReservedName;
  item.index = rn;
  chars_.append(str, length);
}

void Markup::addS(Char c)
{
  if (items_.length() > 0) {
    MarkupItem &item = items_.last();
    if (item.type == Markup::s) {
      item.nChars += 1;
      chars_ += c;
      return;
    }
  }
  MarkupItem &item = items_.grow();
  item.type = Markup::s;
  item.nChars = 1;
  chars_ += c;
}

void Markup::addS(const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::s;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addCommentStart()
{
  MarkupItem &item = items_.grow();
  item.type = Markup::comment;
  item.nChars = 0;
}

void Markup::addRefEndRe()
{
  MarkupItem &item = items_.grow();
  item.type = Markup::refEndRe;
}

void Markup::addCommentChar(Char c)
{
  items_.last().nChars += 1;
  chars_ += c;
}

void Markup::addName(const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::name;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addName(const Char *str, size_t length)
{
  MarkupItem &item = items_.grow();
  item.nChars = length;
  item.type = Markup::name;
  chars_.append(str, length);
}

void Markup::addNumber(const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::number;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addNameToken(const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::nameToken;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addAttributeValue(const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::attributeValue;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addShortref(const InputSource *in)
{
  MarkupItem &item = items_.grow();
  size_t length = in->currentTokenLength();
  item.nChars = length;
  item.type = Markup::shortref;
  chars_.append(in->currentTokenStart(), length);
}

void Markup::addEntityStart(const ResourcePointer<EntityOrigin> &origin)
{
  MarkupItem &item = items_.grow();
  item.type = Markup::entityStart;
  item.origin = new ConstResourcePointer<Origin>(origin.pointer());
}

void Markup::addEntityEnd()
{
  items_.grow().type = Markup::entityEnd;
}

void Markup::addLiteral(const Text &text)
{
  MarkupItem &item = items_.grow();
  item.type = Markup::literal;
  item.text = new Text(text);
}

void Markup::addSdLiteral(const SdText &sdText)
{
  MarkupItem &item = items_.grow();
  item.type = Markup::sdLiteral;
  item.sdText = new SdText(sdText);
}

void Markup::changeToAttributeValue(size_t i)
{
  ASSERT(items_[i].type == Markup::name);
  items_[i].type = Markup::attributeValue;
}

void Markup::moveTo(Markup &to)
{
  chars_.moveTo(to.chars_);
  items_.moveTo(to.items_);
}

MarkupIter::MarkupIter(const Markup &m)
: chars_(m.chars_.pointer()),
  items_(m.items_.pointer()),
  nItems_(m.items_.length()),
  index_(0),
  charIndex_(0)
{
}

void MarkupIter::advance(Location &loc,
			 const ConstResourcePointer<Syntax> &syntax)
{
  switch (items_[index_].type) {
  case Markup::delimiter:
    loc += syntax->delimGeneral(delimGeneral()).length();
    break;
  case Markup::refEndRe:
    loc += 1;
    break;
  case Markup::reservedName:
  case Markup::sdReservedName:
  case Markup::name:
  case Markup::nameToken:
  case Markup::number:
  case Markup::attributeValue:
  case Markup::s:
  case Markup::shortref:
    loc += items_[index_].nChars;
    charIndex_ += items_[index_].nChars;
    break;
  case Markup::comment:
    loc += items_[index_].nChars + (2 * syntax->delimGeneral(Syntax::dCOM).length());
    charIndex_ += items_[index_].nChars;
    break;
  case Markup::entityStart:
    loc = Location(*items_[index_].origin, 0);
    break;
  case Markup::entityEnd:
    {
      ConstResourcePointer<Origin> origin(loc.origin());
      loc = origin->parent();
      loc += origin->refLength();
    }
    break;
  case Markup::literal:
    {
      const Text &text = *items_[index_].text;
      text.endDelimLocation(loc);
      Boolean lita;
      text.delimType(lita);
      loc
	+= syntax->delimGeneral(lita ? Syntax::dLITA : Syntax::dLIT).length();
    }
    break;
  case Markup::sdLiteral:
    {
      const SdText &text = *items_[index_].sdText;
      loc = text.endDelimLocation();
      loc += 1;
    }
    break;
  }
  index_++;
}

void MarkupIter::advance()
{
  switch (items_[index_].type) {
  case Markup::reservedName:
  case Markup::sdReservedName:
  case Markup::name:
  case Markup::nameToken:
  case Markup::number:
  case Markup::attributeValue:
  case Markup::s:
  case Markup::comment:
  case Markup::shortref:
    charIndex_ += items_[index_].nChars;
    break;
  }
  index_++;
}
