// Copyright (c) 1994 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif
#include "config.h"
#include "ParserState.H"
#include "InternalInputSource.H"
#include "MessageArg.H"
#include "macros.H"
#include "SgmlParser.H"
#include "IListIter.H"
#include "setAll.H"
#include "ParserMessages.H"
#include "Undo.H"

const Location ParserState::nullLocation_;
const ShortReferenceMap ParserState::theEmptyMap;

static const size_t eventSizes[] = {
#define EVENT(c, f) sizeof(c),
#include "events.H"
#undef EVENT
};

static const size_t internalSizes[] = {
  sizeof(InternalInputSource),
  sizeof(EntityOrigin),
  sizeof(OpenElement),
  sizeof(UndoStartTag),
  sizeof(UndoEndTag),
  sizeof(UndoTransition)
};

static
size_t maxSize(const size_t *v, size_t n)
{
  size_t max = 0;
  for (size_t i = 0; i < n; i++) {
    if (v[i] > max)
      max = v[i];
  }
  return max;
}

ParserState::ParserState(const ConstResourcePointer<EntityManager> &em,
			 const ConstResourcePointer<EntityCatalog> &ec,
			 const ParserOptions &opt)
: entityManager_(em),
  entityCatalog_(ec),
  options_(opt),
  inInstance_(0),
  keepingMessages_(0),
  eventAllocator_(maxSize(eventSizes, SIZEOF(eventSizes)), 50),
  internalAllocator_(maxSize(internalSizes, SIZEOF(internalSizes)), 50),
  handler_(&eventQueue_),
  subdocLevel_(0),
  documentElementContainer_(CString(), size_t(-1))
{
  init();
}

ParserState::ParserState(const ParserState &parent,
			 const ParserSubdocOptions &options)
: entityManager_(parent.entityManager_),
  entityCatalog_(parent.entityCatalog_),
  options_(parent.options_),
  inInstance_(0),
  keepingMessages_(0),
  eventAllocator_(maxSize(eventSizes, SIZEOF(eventSizes)), 50),
  internalAllocator_(maxSize(internalSizes, SIZEOF(internalSizes)), 50),
  handler_(&eventQueue_),
  subdocLevel_(options.referenced ? parent.subdocLevel_ + 1 : 1),
  documentElementContainer_(CString(), size_t(-1)),
  prologSyntax_(parent.prologSyntax_),
  instanceSyntax_(parent.instanceSyntax_),
  syntax_(parent.prologSyntax_),
  sd_(parent.sd_)
{
  init();
  if (options.inheritActiveLinkTypes)
    activeLinkTypes_ = parent.activeLinkTypes_;
  if (options.inheritIncludes)
    includes_ = parent.includes_;
  for (int i = 0; i < nModes; i++)
    recognizers_[i] = parent.recognizers_[i];
  if (subdocLevel() == sd().subdoc() + 1)
    message(ParserMessages::subdocLevel, NumberMessageArg(sd().subdoc()));
}

void ParserState::init()
{
  inputLevel_ = 0;
  specialParseInputLevel_ = 0;
  markedSectionLevel_ = 0;
  markedSectionSpecialLevel_ = 0;
  currentMode_ = proMode;
  hadLpd_ = 0;
  resultAttributeSpecMode_ = 0;
  pass2_ = 0;
  activeLinkTypesSubsted_ = 0;
  allowPass2_ = 0;
  hadPass2Start_ = 0;
  pcdataRecovering_ = 0;
  currentMarkup_ = 0;
}

void ParserState::allDone()
{
  doFunction_ = 0;
}

void ParserState::setPass2Start()
{
  ASSERT(inputLevel_ == 1);
  if (hadPass2Start_)
    return;
  hadPass2Start_ = 1;
  if (!pass2() && sd().link() && activeLinkTypes_.length() > 0) {
    allowPass2_ = 1;
    pass1Handler_.init(handler_);
    handler_ = &pass1Handler_;
    const InputSourceOrigin *p
      = currentLocation().origin()->asInputSourceOrigin();
    pass2StartOffset_= p->startOffset(currentLocation().index());
  }
  else {
    allowPass2_ = 0;
    currentInput()->willNotRewind();
  }
}

void ParserState::allLinkTypesActivated()
{
  if (activeLinkTypes_.length() == 0 && inputLevel_ == 1)
    currentInput()->willNotRewind();
}

Boolean ParserState::maybeStartPass2()
{
  if (pass2_ || !allowPass2_)
    return 0;
  handler_ = pass1Handler_.origHandler();
  if (!nActiveLink() || pass1Handler_.hadError()) {
    while (!pass1Handler_.empty())
      pass1Handler_.get()->handle(*handler_);
    InputSource *top = 0;
    for (IListIter<InputSource> iter(inputStack_);
	 !iter.done();
	 iter.next())
      top = iter.cur();
    if (top)
      top->willNotRewind();
    return 0;
  }
  pass1Handler_.clear();
  while (inputLevel_ > 1) {
    InputSource *p = inputStack_.get();
    inputLevel_--;
    delete p;
  }
  // Caller will call allDone() if inputLevel_ is 0.
  if (inputLevel_ == 0)
    return 0;
  if (!inputStack_.head()->rewind(*this)) {
    inputLevel_ = 0;
    delete inputStack_.get();
    return 0;
  }
  inputStack_.head()->willNotRewind();
  for (; pass2StartOffset_ > 0; pass2StartOffset_--)
    if (inputStack_.head()->get(messenger()) == InputSource::eE) {
      message(ParserMessages::pass2Ee);
      inputLevel_ = 0;
      delete inputStack_.get();
      return 0;
    }
  init();
  inputLevel_ = 1;
  inInstance_ = 0;
  defDtd_.clear();
  defLpd_.clear();
  dtd_[0].moveTo(pass1Dtd_);
  dtd_.clear();
  dsEntity_.clear();
  currentDtd_.clear();
  doFunction_ = 0;
  pass2_ = 1;
  lpd_.clear();
  allLpd_.clear();
  return 1;
}

ConstResourcePointer<AttributeValue> ParserState::makeImpliedAttributeValue()
{
  if (impliedAttributeValue_.isNull())
    impliedAttributeValue_ = new ImpliedAttributeValue;
  return impliedAttributeValue_;
}

Boolean ParserState::referenceDsEntity(const Location &loc)
{
  if (dsEntity_.isNull())
    return 0;
  ResourcePointer<EntityOrigin> origin
    = new (internalAllocator()) EntityOrigin(dsEntity_, loc);
  dsEntity_->dsReference(*this, origin);
  dsEntity_.clear();
  return inputLevel() > 1;
}

void ParserState::startDtd(const CString &name)
{
  defDtd_ = new Dtd(name, dtd_.length() == 0);
  for (size_t i = 0; i < includes_.length(); i++) {
    CString name = includes_[i];
    const SubstTable<Char> *subst = syntax().entitySubstTable();
    for (size_t j = 0; j < name.length(); j++)
      subst->subst(name[j]);
    Text text;
    text.addChars(syntax().reservedName(Syntax::rINCLUDE), Location());
    Entity *entity
      = new InternalTextEntity(name,
			       Entity::parameterEntity,
			       Location(),
			       text,
			       InternalTextEntity::none);
    entity->setUsed();
    defDtd_->insertEntity(entity);
  }
  currentDtd_ = defDtd_;
  currentMode_ = dsMode;
}

void ParserState::addInclude(const CString &str)
{
  includes_.grow() = str;
}

void ParserState::endDtd()
{
  dtd_.grow() = defDtd_;
  defDtd_.clear();
  currentDtd_.clear();
  currentMode_ = proMode;
}

void ParserState::startLpd(ResourcePointer<Lpd> &lpd)
{
  defLpd_ = lpd;
  currentDtd_ = defLpd_->sourceDtd();
  currentMode_ = dsMode;
}

void ParserState::endLpd()
{
  hadLpd_ = 1;
  if (defLpd_->active())
    lpd_.grow() = defLpd_;
  allLpd_.grow() = defLpd_;
  defLpd_.clear();
  currentDtd_.clear();
  currentMode_ = proMode;
}

void ParserState::popInputStack()
{
  ASSERT(inputLevel_ > 0);
  InputSource *p = inputStack_.get();
  inputLevel_--;
  delete p;
  if (specialParseInputLevel_ > 0 && inputLevel_ == specialParseInputLevel_)
    currentMode_ = specialParseMode_;
  if (currentMode_ == dsiMode
      && inputLevel_ == 1
      && markedSectionLevel_ == 0)
    currentMode_ = dsMode;
}

void ParserState::setSyntax(Syntax *syntax)
{
  syntax_ = syntax;
  prologSyntax_ = syntax;
  instanceSyntax_ = syntax;
}

void ParserState::setSyntaxes(Syntax *prologSyntax, Syntax *instanceSyntax)
{
  syntax_ = prologSyntax;
  prologSyntax_ = prologSyntax;
  instanceSyntax_ = instanceSyntax;
}

void ParserState::pushInput(InputSource *in)
{
  if (!in)
    return;
  if (!syntax_.isNull() && syntax_->multicode())
    in->setMarkupScanTable(syntax_->markupScanTable());
  inputStack_.insert(in);
  inputLevel_++;
  if (specialParseInputLevel_ > 0 && inputLevel_ > specialParseInputLevel_)
    currentMode_ = rcconeMode;	// mode for rcdata in an entity
  else if (currentMode_ == dsMode)
    currentMode_ = dsiMode;
}

void ParserState::startMarkedSection(const Location &loc)
{
  markedSectionLevel_++;
  markedSectionStartLocation_.grow() = loc;
  if (currentMode_ == dsMode)
    currentMode_ = dsiMode;
  if (markedSectionSpecialLevel_)
    markedSectionSpecialLevel_++;
}

void ParserState::startSpecialMarkedSection(Mode mode, const Location &loc)
{
  markedSectionLevel_++;
  markedSectionStartLocation_.grow() = loc;
  specialParseInputLevel_ = inputLevel_;
  markedSectionSpecialLevel_ = 1;
  specialParseMode_ = currentMode_ = mode;
}

void ParserState::endMarkedSection()
{
  ASSERT(markedSectionLevel_ > 0);
  markedSectionLevel_--;
  markedSectionStartLocation_.setLength(markedSectionStartLocation_.length()
					- 1);
  if (markedSectionSpecialLevel_ > 0) {
    markedSectionSpecialLevel_--;
    if (markedSectionSpecialLevel_ > 0)
      return;			// remain in imsMode
    specialParseInputLevel_ = 0;
    if (inInstance_)
      currentMode_ = openElements_.head()->mode(netEnablingCount_ > 0);
    else
      currentMode_ = dsiMode;
  }
  if (currentMode_ == dsiMode
      && inputLevel_ == 1
      && markedSectionLevel_ == 0)
    currentMode_ = dsMode;
}

void ParserState::pushElement(OpenElement *e)
{
  tagLevel_++;
  openElementCount_[e->type()->index()]++;
  const ElementDefinition *def = e->type()->definition();
  if (def) {
    size_t i;
    for (i = 0; i < def->nInclusions(); i++)
      includeCount_[def->inclusion(i)->index()]++;
    for (i = 0; i < def->nExclusions(); i++) {
      excludeCount_[def->exclusion(i)->index()]++;
      totalExcludeCount_++;
    }
  }
  if (e->netEnabling())
    netEnablingCount_++;
  // the start tag of this element may have been implied by data
  // inside a cdata or rcdata marked section
  if (markedSectionSpecialLevel_ == 0) {
    currentMode_ = e->mode(netEnablingCount_ > 0);
    if (e->requiresSpecialParse()) {
      specialParseMode_ = currentMode_;
      specialParseInputLevel_ = inputLevel_;
    }
  }
  openElements_.insert(e);
  pcdataRecovering_ = 0;
}

// PCDATA was encountered somewhere where it was not allowed.
// Change the current mode to improve recovery.

void ParserState::pcdataRecover()
{
  switch (currentMode_) {
  case econMode:
    currentMode_ = mconMode;
    break;
  case econnetMode:
    currentMode_ = mconnetMode;
    break;
  default:
    break;
  }
  pcdataRecovering_ = 1;
}

OpenElement *ParserState::popSaveElement()
{
  ASSERT(tagLevel_ > 0);
  OpenElement *e = openElements_.get();
  tagLevel_--;
  openElementCount_[e->type()->index()]--;
  const ElementDefinition *def = e->type()->definition();
  if (def) {
    size_t i;
    for (i = 0; i < def->nInclusions(); i++)
      includeCount_[def->inclusion(i)->index()]--;
    for (i = 0; i < def->nExclusions(); i++) {
      excludeCount_[def->exclusion(i)->index()]--;
      totalExcludeCount_--;
    }
  }
  if (e->netEnabling())
    netEnablingCount_--;
  // the end tag of this element may have been implied by data
  // inside a cdata or rcdata marked section
  if (markedSectionSpecialLevel_ == 0) {
    currentMode_ = currentElement().mode(netEnablingCount_ > 0);
    specialParseInputLevel_ = 0;
  }
  lastEndedElementType_ = e->type();
  pcdataRecovering_ = 0;
  return e;
}

void ParserState::popElement()
{
  delete popSaveElement();
}
			      
Boolean ParserState::entityIsOpen(const Entity *entity) const
{
  for (IListIter<InputSource> iter(inputStack_); !iter.done(); iter.next()) {
    const EntityOrigin *eo
      = iter.cur()->currentLocation().origin()->asEntityOrigin();
    if (eo && eo->entity().pointer() == entity)
      return 1;
  }
  return 0;
}

void ParserState::startInstance()
{
  if (!instanceSyntax_.isNull())
    syntax_ = instanceSyntax_;
  currentMode_ = econMode;
  currentDtd_ = dtd_[0];
  Vector<Owner<ContentToken> > tokens(1);
  tokens[0] = new ElementToken(currentDtd_->documentElementType(),
			       ContentToken::none);
  Owner<ModelGroup> model(new SeqModelGroup(tokens, ContentToken::none));
  Owner<CompiledModelGroup> compiledModel(new CompiledModelGroup(model));
  GrowableVector<ContentModelAmbiguity> ambiguities;
  Boolean pcdataUnreachable;
  compiledModel->compile(currentDtd().nElementTypeIndex(), ambiguities,
			 pcdataUnreachable);
  ASSERT(ambiguities.length() == 0);
  ConstResourcePointer<ElementDefinition> def
    = new ElementDefinition(Location(),
			    0,
			    0,
			    ElementDefinition::modelGroup,
			    compiledModel);
  documentElementContainer_.setElementDefinition(def, 0);
  tagLevel_ = 0;
  while (!openElements_.empty())
    delete openElements_.get();
  openElements_.insert(new (internalAllocator_)
		         OpenElement(&documentElementContainer_,
				     0,
				     0,
				     &theEmptyMap,
				     Location()));
  inInstance_ = 1;
  if (sd().rank()) {
    CString empty;
    currentRank_.init(currentDtd().nRankStem());
    setAll(currentRank_.pointer(), currentRank_.length(), empty);
  }
  includeCount_.setLength(currentDtd().nElementTypeIndex());
  setAll(includeCount_.pointer(), includeCount_.length(), unsigned(0));
  excludeCount_.setLength(currentDtd().nElementTypeIndex());
  setAll(excludeCount_.pointer(), excludeCount_.length(), unsigned(0));
  openElementCount_.setLength(currentDtd().nElementTypeIndex());
  setAll(openElementCount_.pointer(), openElementCount_.length(),
	 unsigned(0));
  netEnablingCount_ = 0;
  totalExcludeCount_ = 0;
  lastEndedElementType_ = 0;
  undefinedElementTypeTable_.clear();
  currentAttributes_.init(currentDtd().nCurrentAttribute());
  idTable_.clear();
}

const ElementType *
ParserState::lookupCreateUndefinedElement(const CString &name)
{
  const ElementType *e = undefinedElementTypeTable_.lookup(name);
  if (e)
    return e;
  ElementType *p = new ElementType(name,
				   currentDtd_->nElementTypeIndex()
				   + undefinedElementTypeTable_.count());
  p->setElementDefinition(new ElementDefinition(currentLocation(),
						ElementDefinition::undefinedIndex,
						(ElementDefinition::omitStart
						 |ElementDefinition::omitEnd),
						ElementDefinition::any),
			  0);
  undefinedElementTypeTable_.insert(p);
  includeCount_.grow() = 0;
  excludeCount_.grow() = 0;
  openElementCount_.grow() = 0;
  return p;
}

#ifdef __GNUG__
typedef IListIter<OpenElement> Dummy_IListIter_OpenElement;
#endif

void ParserState::getOpenElementInfo(Vector<OpenElementInfo> &v) const
{
  if (!inInstance())
    return;
  v.init(tagLevel_);
  unsigned i = tagLevel_;
  for (IListIter<OpenElement> iter(openElements_);
       !iter.done() && i > 0;
       iter.next()) {
    OpenElementInfo &e = v[--i];
    e.gi = iter.cur()->type()->name();
    const LeafContentToken *token = iter.cur()->currentPosition();
    if (token && !token->isInitial()) {
      e.matchIndex = token->typeIndex() + 1;
      const ElementType *type = token->elementType();
      if (type)
	e.matchType = type->name();
      else {
	e.matchType = syntax().delimGeneral(Syntax::dRNI);
	e.matchType += syntax().reservedName(Syntax::rPCDATA);
      }
    }
    e.included = iter.cur()->included();
  }
}

Boolean ParserState::checkImplyLoop(unsigned count)
{
  for (IListIter<OpenElement> iter(openElements_);
       count > 0;
       iter.next(), count--)
    if (iter.cur()->type() == openElements_.head()->type()
	// I'm not sure whether this is necessary.
	&& iter.cur()->matchState() == openElements_.head()->matchState())
      return 0;
  return 1;
}

Id *ParserState::lookupCreateId(const CString &name)
{
  Id *id = idTable_.lookup(name);
  if (!id) {
    id = new Id(name);
    idTable_.insert(id);
  }
  return id;
}

ConstResourcePointer<Entity>
ParserState::lookupEntity(Boolean isParameter,
			  const CString &name,
			  const Location &useLocation,
			  Boolean referenced,
			  Boolean &defaulted)
{
  defaulted = 0;
  Dtd *dtd;
  if (resultAttributeSpecMode_)
    dtd = defComplexLpd().resultDtd().pointer();
  else
    dtd = currentDtd_.pointer();
  if (dtd) {
    ResourcePointer<Entity> entity(dtd->lookupEntity(isParameter, name));
    // Did we find it in pass1Dtd?
    // Did we look at the defaultEntity?
    if (!inInstance_ && pass2() && dtd->isBase()
	&& !resultAttributeSpecMode_
	&& (entity.isNull() || entity->declInLpdNamePointer() == 0)) {
      ConstResourcePointer<Entity> entity1
	= pass1Dtd_->lookupEntity(isParameter, name);
      if (!entity1.isNull() && entity1->declInLpdNamePointer() != 0) {
	if (referenced)
	  noteReferencedEntity(entity1, 1, 0);
	return entity1;
      }
      else if (!entity.isNull()) {
	if (referenced)
	  noteReferencedEntity(entity, 0, 0);
	entity->setUsed();
	return entity;
      }
    }
    else if (!entity.isNull()) {
      entity->setUsed();
      return entity;
    }
    if (!isParameter) {
      ConstResourcePointer<Entity> entity(dtd->defaultEntity());
      Boolean note = 0;
      Boolean usedPass1 = 0;
      if (!inInstance_ && pass2() && dtd->isBase()
	  && !resultAttributeSpecMode_
	  && (entity.isNull() || entity->declInLpdNamePointer() == 0)) {
	if (referenced)
	  note = 1;
	ConstResourcePointer<Entity> entity1 = pass1Dtd_->defaultEntity();
	if (!entity1.isNull() && entity1->declInLpdNamePointer() != 0) {
	  usedPass1 = 1;
	  entity = entity1;
	}
      }
      if (!entity.isNull()) {
	ConstResourcePointer<Entity> tem(dtd->lookupDefaultedEntity(name));
	if (!tem.isNull())
	  entity = tem;
	else {
	  ResourcePointer<Entity> p(entity->copy());
	  p->setName(name);
	  p->generateSystemId(*this);
	  dtd->insertDefaultedEntity(p);
	  entity = p;
	  if (inInstance_)
	    eventHandler().entityDefaulted(new (eventAllocator())
					   EntityDefaultedEvent(entity,
								useLocation));
	}
	if (note)
	  noteReferencedEntity(entity, usedPass1, 1);
	defaulted = 1;
      }
      return entity;
    }
  }
  return (Entity *)0;
}

void ParserState::noteReferencedEntity(const ConstResourcePointer<Entity> &entity,
				       Boolean foundInPass1Dtd,
				       Boolean lookedAtDefault)
{
  LpdEntityRef ref;
  ref.entity = entity;
  ref.lookedAtDefault = lookedAtDefault;
  ref.foundInPass1Dtd = foundInPass1Dtd;
  LpdEntityRef *old = lpdEntityRefs_.lookup(ref);
  if (!old)
    lpdEntityRefs_.insert(new LpdEntityRef(ref));
}

// Compare entity definitions.
// e1 is the original (will not be an external non-text entity).
// FIXME should look at generated sysids as well.
static
Boolean sameEntityDef(const Entity *e1, const Entity *e2)
{
  if (e1->dataType() != e2->dataType())
    return 0;
  const InternalEntity *i1 = e1->asInternalEntity();
  const InternalEntity *i2 = e2->asInternalEntity();
  if (i1) {
    if (!i2)
      return 0;
    if (i1->string() != i2->string())
      return 0;
    return 1;
  }
  else if (i2)
    return 0;
  const ExternalEntity *x1 = e1->asExternalEntity();
  const ExternalEntity *x2 = e2->asExternalEntity();
  const CString *s1 = x1->externalId().systemIdString();
  const CString *s2 = x2->externalId().systemIdString();
  if (s1) {
    if (!s2)
      return 0;
    if (*s1 != *s2)
      return 0;
  }
  else if (s2)
    return 0;
  s1 = x1->externalId().publicIdString();
  s2 = x2->externalId().publicIdString();
  if (s1) {
    if (!s2)
      return 0;
    if (*s1 != *s2)
      return 0;
  }
  else if (s2)
    return 0;
  return 1;
}

void ParserState::checkEntityStability()
{
  LpdEntityRefSetIter iter(lpdEntityRefs_);
  LpdEntityRef *ref;
  while ((ref = iter.next()) != 0) {
    ConstResourcePointer<Entity> entity
      = dtd_[0]->lookupEntity(ref->entity->declType()
			      == Entity::parameterEntity,
			      ref->entity->name());
    if (entity.isNull() && ref->lookedAtDefault)
      entity = dtd_[0]->defaultEntity();
    if (entity.isNull()
	? ref->foundInPass1Dtd
	: !sameEntityDef(ref->entity.pointer(), entity.pointer()))
      message(((ref->entity->declType()
		== Entity::parameterEntity)
	       ? ParserMessages::unstableLpdParameterEntity
	       : ParserMessages::unstableLpdGeneralEntity),
	      StringMessageArg(ref->entity->name()));
  }
  lpdEntityRefs_.clear();
}    

Boolean ParserState::appendCurrentRank(CString &str, const RankStem *stem)
     const
{
  const CString &suffix = currentRank_[stem->index()];
  if (suffix.length() > 0) {
    str += suffix;
    return 1;
  }
  return 0;
}

void ParserState::setCurrentRank(const RankStem *stem, const CString &suffix)
{
  currentRank_[stem->index()] = suffix;
}

void ParserState::getCurrentToken(const SubstTable<Char> *subst,
				  CString &str) const
{
  InputSource *in = currentInput();
  const Char *p = in->currentTokenStart();
  size_t count = in->currentTokenLength();
  str.init(count);
  Char *s = str.pointer();
  for (; count > 0; --count)
    *s++ = (*subst)[*p++];
}

void ParserState::queueMessage(MessageEvent *event)
{
  if (keepingMessages_)
    keptMessages_.append(event);
  else
    handler_->message(event);
}

void ParserState::releaseKeptMessages()
{
  keepingMessages_ = 0;
  while (!keptMessages_.empty())
    handler_->message(keptMessages_.get());
}

void ParserState::discardKeptMessages()
{
  keepingMessages_ = 0;
  keptMessages_.clear();
}

void ParserState::initMessage(Message &msg)
{
  getOpenElementInfo(msg.openElementInfo);
  msg.loc = currentLocation();
}

void ParserState::dispatchMessage(Message &msg)
{
  queueMessage(new MessageEvent(msg));
}

AttributeList *
ParserState::allocAttributeList(const ConstResourcePointer<AttributeDefinitionList> &def,
				unsigned i)
{
  if (i < attributeLists_.length())
    attributeLists_[i]->init(def);
  else {
    attributeLists_.setLength(i + 1);
    attributeLists_[i] = new AttributeList(def);
  }
  return attributeLists_[i].pointer();
}

void ParserState::activateLinkType(const CString &name)
{
  if (!hadPass2Start_ && !pass2_)
    activeLinkTypes_.grow() = name;
  else
    message(ParserMessages::linkActivateTooLate);
}

Boolean ParserState::shouldActivateLink(const CString &name) const
{
  if (!activeLinkTypesSubsted_) {
    // FIXME use mutable
    ParserState *state = (ParserState *)this;
    for (size_t i = 0; i < activeLinkTypes_.length(); i++)
      for (size_t j = 0; j < activeLinkTypes_[i].length(); j++)
	syntax().generalSubstTable()->subst(state->activeLinkTypes_[i][j]);
    state->activeLinkTypesSubsted_ = 1;
  }
  for (size_t i = 0; i < activeLinkTypes_.length(); i++)
    if (name == activeLinkTypes_[i])
      return 1;
  return 0;
}

ResourcePointer<Dtd> ParserState::lookupDtd(const CString &name)
{
  for (size_t i = 0; i < dtd_.length(); i++)
    if (dtd_[i]->name() == name)
      return dtd_[i];
  return ResourcePointer<Dtd>();
}

ConstResourcePointer<Lpd> ParserState::lookupLpd(const CString &name) const
{
  for (size_t i = 0; i < allLpd_.length(); i++)
    if (allLpd_[i]->name() == name)
      return allLpd_[i];
  return ConstResourcePointer<Lpd>();
}
