// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "CopyEventHandler.H"
#include "macros.H"

inline
OutputCharStream &operator<<(OutputCharStream &os, const MarkupIter &iter)
{
  return os.write(iter.charsPointer(), iter.charsLength());
}

#ifdef __GNUG__
inline
#endif
Boolean CopyEventHandler::noOutput()
{
  if (inInstance_) {
    if (normalizeFlags_ & normalizeExpand) {
      if (entityLevel_ >= outputEntityLevel_)
	return 0;
    }
    else {
      if (entityLevel_ == outputEntityLevel_)
	return 0;
    }
  }
  else if (normalizeFlags_ & normalizeIncludeProlog) {
    if (normalizeFlags_ & normalizeExpandProlog) {
      if (entityLevel_ >= outputEntityLevel_)
	return 0;
    }
    else {
      if (entityLevel_ == outputEntityLevel_)
	return 0;
    }
  }
  return 1;
}

inline
Boolean CopyEventHandler::doNothing(Event *event)
{
  if (noOutput()) {
    delete event;
    return 1;
  }
  else
    return 0;
}


inline
void CopyEventHandler::withNamedCharRef(const CString &str,
					const Location &loc)
{
  withNamedCharRef(str.pointer(), str.length(), loc);
}


CopyEventHandler::CopyEventHandler(OutputCharStream *os,
				   unsigned normalizeFlags,
				   const CString &outputEntity)
: os_(os), topOs_(os), inInstance_(0), entityLevel_(0),
  normalizeFlags_(normalizeFlags), outputEntity_(outputEntity),
  omittagHoist_(0), inSpecialMarkedSection_(0),
  currentAttributes_(0)
{
  outputEntityLevel_ = outputEntity_.length() == 0 ? 0 : unsigned(-1);
}

void CopyEventHandler::markup(const Location &loc,
			      const Markup &markup)
{
  if (!noOutput())
    outputMarkup(loc, markup);
}

void CopyEventHandler::sgmlDecl(SgmlDeclEvent *event)
{
  if (event->inDocumentEntity()
      && (normalizeFlags_ & normalizeIncludeProlog)) {
    syntax_ = event->refSyntaxPointer();
    sd_ = event->refSdPointer();
    outputMarkup(event->location(), event->markup());
  }
  syntax_ = event->prologSyntaxPointer();
  instanceSyntax_ = event->instanceSyntaxPointer();
  if (instanceSyntax_->namecaseGeneral())
    instanceSyntax_->generalSubstTable()->inverseTable(lowerSubst_);
  else if (instanceSyntax_->namecaseEntity())
    instanceSyntax_->entitySubstTable()->inverseTable(lowerSubst_);
  sd_ = event->sdPointer();
  delete event;
}

void CopyEventHandler::endProlog(EndPrologEvent *event)
{
  inInstance_ = 1;
  syntax_ = instanceSyntax_;
  delete event;
}

void CopyEventHandler::data(DataEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  if (event->entity())
    entityRef(event->location().origin()->asEntityOrigin());
  else {
    size_t n = event->dataLength();
    unsigned long dummy;
    if (n > 1 || !event->isRe(dummy))
      writeData(event->data(), n, event->location());
  }
  delete event;
}

void CopyEventHandler::writeData(const Char *p, size_t n,
				 const Location &loc)
{
  const Markup *markupPtr;
  if (n == 1 && loc.origin()->isNumericCharRef(markupPtr)) {
    if (markupPtr)
      outputMarkup(loc.origin()->parent(), *markupPtr);
  }
  else
    withNamedCharRef(p, n, loc);
}

void CopyEventHandler::withNamedCharRef(const Char *p, size_t n,
					const Location &loc)
{
  if (n > 0) {
    const Origin *origin = loc.origin().pointer();
    if (origin) {
      NamedCharRef ref;
      Index startIndex;
      if (origin->isNamedCharRef(loc.index(), ref)) {
	Markup markup;
	markup.addDelim(Syntax::dCRO);
	markup.addName(ref.origName().pointer(), ref.origName().length());
	switch (ref.refEndType()) {
	case NamedCharRef::endOmitted:
	  break;
	case NamedCharRef::endRE:
	  markup.addRefEndRe();
	  break;
	case NamedCharRef::endRefc:
	  markup.addDelim(Syntax::dREFC);
	  break;
	}
	outputMarkup(Location(loc.origin(), ref.refStartIndex()), markup);
	p++;
	n--;
      }
    }
  }
  os().write(p, n);
}

void CopyEventHandler::reOrigin(ReOriginEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  Char c = event->re();
  withNamedCharRef(&c, 1, event->location());
  delete event;
}

void CopyEventHandler::sSep(SSepEvent *event)
{
  if (doNothing(event))
    return;
  withNamedCharRef(event->data(), event->dataLength(), event->location());
  delete event;
}

void CopyEventHandler::ignoredRs(IgnoredRsEvent *event)
{
  if (doNothing(event))
    return;
  Char c = event->rs();
  withNamedCharRef(&c, 1, event->location());
  delete event;
}

void CopyEventHandler::startElement(StartElementEvent *event)
{
  mustOmitEnd_ = event->mustOmitEnd();
  const Markup *markup = event->markupPtr();
  if (!markup) {
    if (normalizeFlags_ & normalizeExpand) {
      if (outputEntityLevel_ > entityLevel_ - omittagHoist_) {
	delete event;
	return;
      }
      if (omittagHoist_ >= entityStack_.length())
	os_ = topOs_;
    }
    else if (entityLevel_ - omittagHoist_ != outputEntityLevel_) {
      delete event;
      return;
    }
  }
  else {
    omittagHoist_ = 0;
    if (doNothing(event))
      return;
  }
  if (normalizeFlags_ & normalizeExpandAll)
    handleChange();
  if (markup) {
    Boolean hadName = 0;
    Boolean closed = 1;
    MarkupIter iter(*markup);
    while (iter.valid()) {
      switch (iter.type()) {
      case Markup::delimiter:
	switch (iter.delimGeneral()) {
	case Syntax::dTAGC:
	  closed = 1;
	  if (!hadName) {
	    CString nameBuf;
	    CString tag(elementTypeOrigName(event->elementType(), nameBuf));
	    if (normalizeFlags_ & normalizeEmptytag) {
	      handleChange();
	      os() << tag;
	      tag.clear();
	    }
	    unspecifiedAttributeValues(event->attributes(), tag);
	  }
	  os() << syntax_->delimGeneral(iter.delimGeneral());
	  break;
	case Syntax::dNET:
	  closed = 1;
	  if (normalizeFlags_ & normalizeNet) {
	    handleChange();
	    os() << syntax_->delimGeneral(Syntax::dTAGC);
	    break;
	  }
	  // fall through
	default:
	  os() << syntax_->delimGeneral(iter.delimGeneral());
	  break;
	}
	iter.advance();
	break;
      case Markup::name:
	{
	  ASSERT(!hadName);
	  const ElementType *elementType = event->elementType();
	  if (elementType->index() >= elementTypeOrigNames_.length())
	    elementTypeOrigNames_.setLength(elementType->index() + 1);
	  CString &elementTypeOrigName
	    = elementTypeOrigNames_[elementType->index()];
	  if (elementTypeOrigName.length() == 0) {
	    elementTypeOrigName.set(iter.charsPointer(), iter.charsLength());
	    // add rank if missing
	    elementTypeOrigName.append(event->name().pointer() +
				       elementTypeOrigName.length(),
				       event->name().length()
				       - elementTypeOrigName.length());
	  }
	  os() << iter;
	  if (normalizeFlags_ & normalizeRank) {
	    for (size_t i = iter.charsLength();
		 i < event->name().length();
		 i++) {
	      handleChange();
	      os().put(event->name()[i]);
	    }
	  }
	  attributeSpecList(iter, event->attributes());
	  hadName = 1;
	}
	break;
      case Markup::s:
	os() << iter;
	iter.advance();
	break;
      default:
	CANNOT_HAPPEN();
      }
    }
    if (!closed && (normalizeFlags_ && normalizeUnclosed)) {
      handleChange();
      os() << syntax_->delimGeneral(Syntax::dTAGC);
    }
  }
  else if (normalizeFlags_ & normalizeOmittag) {
    if (inSpecialMarkedSection_) {
      reportTagInSpecialMarkedSection(event->location());
      return;
    }
    handleChange();
    CString nameBuf;
    os() << syntax_->delimGeneral(Syntax::dSTAGO)
      << elementTypeOrigName(event->elementType(), nameBuf);
    unspecifiedAttributeValues(event->attributes(), CString());
    os() << syntax_->delimGeneral(Syntax::dTAGC);
  }
  delete event;
  if (entityStack_.length() > 0 && os_ == topOs_)
    os_ = &entityStack_.last().str;
}

void CopyEventHandler::attributeSpecList(MarkupIter &iter,
					 const AttributeList &atts)
{
  size_t nAtt = atts.length();
  unsigned i;
  unsigned *attIndex;
  if (atts.nSpec()) {
    attIndex = new unsigned[atts.nSpec()];
    for (i = 0; i < atts.nSpec(); i++)
      attIndex[i] = unsigned(-1);
    for (i = 0; i < nAtt; i++)
      if (atts.specified(i))
	attIndex[atts.specIndex(i)] = i;
  }
  else
    attIndex = 0;
  Boolean hadAttname = 0;
  i = 0;
  CString nameBuf;
  for (iter.advance(); iter.valid(); iter.advance())
    switch (iter.type()) {
    case Markup::name:
      os() << iter;
      hadAttname = 1;
      break;
    case Markup::s:
      os() << iter;
      break;
    case Markup::attributeValue:
      if (!hadAttname
	  && attIndex
	  && attIndex[i] != unsigned(-1)
	  && (normalizeFlags_ & (normalizeAttname | normalizeAttvalue))) {
	handleChange();
	os() << generalName(atts.name(attIndex[i]), nameBuf)
	  << syntax_->delimGeneral(Syntax::dVI);
      }
      if (normalizeFlags_ & normalizeAttvalue) {
	handleChange();
	os() << syntax_->delimGeneral(Syntax::dLIT)
	  << iter
	    << syntax_->delimGeneral(Syntax::dLIT);
      }
      else
	os() << iter;
      hadAttname = 0;
      i++;
      break;
    case Markup::literal:
      literal(iter.text());
      i++;
      hadAttname = 0;
      break;
    case Markup::delimiter:
      if (iter.delimGeneral() == Syntax::dVI)
	os() << syntax_->delimGeneral(iter.delimGeneral());
      else {
	unspecifiedAttributeValues(atts, CString());
	delete [] attIndex;
	return;
      }
      break;
    default:
      CANNOT_HAPPEN();
    }
}

void CopyEventHandler::unspecifiedAttributeValues(const AttributeList &atts,
						  const CString &beforeFirst)
{
  if (normalizeFlags_ & (normalizeCurrent|normalizeAttspec)) {
    Boolean first = 1;
    size_t nAtt = atts.length();
    CString nameBuf;
    for (unsigned i = 0; i < nAtt; i++) {
      const Text *text;
      if (!atts.specified(i)
	  && ((normalizeFlags_ & normalizeAttspec)
	      || atts.current(i))
	  && atts.value(i)
	  && atts.value(i)->text(text)) {
	if (first) {
	  handleChange();
	  os() << beforeFirst;
	  first = 0;
	}
	os().put(syntax_->standardFunction(Syntax::fSPACE));
	os() << generalName(atts.name(i), nameBuf)
	  << syntax_->delimGeneral(Syntax::dVI);
	literal(*text);
      }
    }
  }
}

void CopyEventHandler::literal(const Text &text)
{
  TextIter iter(text);
  TextItem::Type type;
  const Char *p;
  size_t n;
  Location loc;
  CString delim;
  Boolean lita;
  if (!text.delimType(lita))
    CANNOT_HAPPEN();
  delim = syntax_->delimGeneral(lita ? Syntax::dLITA : Syntax::dLIT);
  os() << delim;
  int level = 0;
  while (iter.next(type, p, n, loc)) {
    switch (type) {
    case TextItem::ignore:
    case TextItem::data:
      if (!level) {
	Char orig;
	if (n == 1 && loc.origin()->origChar(orig))
	  writeData(&orig, 1, loc.origin()->parent());
	else
	  writeData(p, n, loc);
      }
      break;
    case TextItem::entityStart:
    case TextItem::cdataEntityStart:
    case TextItem::sdataEntityStart:
      if (!level++)
	entityRef(loc.origin()->asEntityOrigin());
      break;
    case TextItem::entityEnd:
    case TextItem::cdataEntityEnd:
    case TextItem::sdataEntityEnd:
      level--;
      break;
    case TextItem::startDelim:
    case TextItem::endDelim:
    case TextItem::endDelimA:
      break;
    }
  }
  if (!text.endDelimLocation(loc))
    CANNOT_HAPPEN();
  withNamedCharRef(delim, loc);
}

void CopyEventHandler::endElement(EndElementEvent *event)
{
  if (mustOmitEnd_) {
    delete event;
    mustOmitEnd_ = 0;
    return;
  }
  const Markup *markup = event->markupPtr();
  if (!markup) {
    if (normalizeFlags_ & normalizeExpand) {
      if (outputEntityLevel_ > entityLevel_ - omittagHoist_) {
	delete event;
	return;
      }
      if (omittagHoist_ >= entityStack_.length())
	os_ = topOs_;
    }
    else if (entityLevel_ - omittagHoist_ != outputEntityLevel_) {
      delete event;
      return;
    }
  }
  else {
    omittagHoist_ = 0;
    if (doNothing(event))
      return;
  }
  if (normalizeFlags_ & normalizeExpandAll)
    handleChange();
  if (markup) {
    Boolean closed = 0;
    Boolean hadAttname = 0;
    for (MarkupIter iter(*markup); iter.valid(); iter.advance())
      switch (iter.type()) {
      case Markup::s:
	os() << iter;
	break;
      case Markup::name:
	{
	  os() << iter;
	  for (size_t i = iter.charsLength();
	       i < event->name().length();
	       i++) {
	    handleChange();
	    os().put(event->name()[i]);
	  }
	  hadAttname = 1;
	}
	break;
      case Markup::delimiter:
	if (iter.delimGeneral() == Syntax::dTAGC) {
	  closed = 1;
	  if (!hadAttname
	      && (normalizeFlags_ & normalizeEmptytag)) {
	    handleChange();
	    CString nameBuf;
	    os() << elementTypeOrigName(event->elementType(), nameBuf);
	  }
	}
	else if (iter.delimGeneral() == Syntax::dNET
		 && (normalizeFlags_ & normalizeNet)) {
	  handleChange();
	  CString nameBuf;
	  os() << syntax_->delimGeneral(Syntax::dETAGO)
	    << elementTypeOrigName(event->elementType(), nameBuf)
	      << syntax_->delimGeneral(Syntax::dTAGC);
	  closed = 1;
	  break;
	}
	os() << syntax_->delimGeneral(iter.delimGeneral());
	break;
      default:
	CANNOT_HAPPEN();
      }
    if (!closed && (normalizeFlags_ && normalizeUnclosed)) {
      handleChange();
      os() << syntax_->delimGeneral(Syntax::dTAGC);
    }
  }
  else if (normalizeFlags_ & normalizeOmittag) {
    if (inSpecialMarkedSection_) {
      reportTagInSpecialMarkedSection(event->location());
      return;
    }
    handleChange();
    CString nameBuf;
    os() << syntax_->delimGeneral(Syntax::dETAGO)
      << elementTypeOrigName(event->elementType(), nameBuf)
	<< syntax_->delimGeneral(Syntax::dTAGC);
  }
  delete event;
  if (entityStack_.length() > 0 && os_ == topOs_)
    os_ = &entityStack_.last().str;
}

void CopyEventHandler::pi(PiEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  if (event->entity())
    entityRef(event->location().origin()->asEntityOrigin());
  else {
    os() << syntax_->delimGeneral(Syntax::dPIO);
    os().write(event->data(), event->dataLength());
    os() << syntax_->delimGeneral(Syntax::dPIC);
  }
  delete event;
}

void CopyEventHandler::sdataEntity(SdataEntityEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  entityRef(event->location().origin()->asEntityOrigin());
  delete event;
}

void CopyEventHandler::externalDataEntity(ExternalDataEntityEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  entityRef(event->entityOrigin().pointer());
  delete event;
}

void CopyEventHandler::subdocEntity(SubdocEntityEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  entityRef(event->entityOrigin().pointer());
  delete event;
}

void CopyEventHandler::markedSectionStart(MarkedSectionStartEvent *event)
{
  omittagHoist_ = 0;
  switch (event->status()) {
  case MarkedSectionEvent::rcdata:
  case MarkedSectionEvent::cdata:
    inSpecialMarkedSection_ = 1;
    break;
  default:
    break;
  }
  if (doNothing(event))
    return;
  if (!(normalizeFlags_ & normalizeMarkedSection)
      || (inInstance_ && inSpecialMarkedSection_))
    outputMarkup(event->location(), event->markup());
  else if (inInstance_ && event->status() != MarkedSectionEvent::ignore) {
    // Put an empty comment so that REs aren't changed.
    // With an ignored marked section, sufficent to have comment at the end.
    handleChange();
    os() << syntax_->delimGeneral(Syntax::dMDO)
         << syntax_->delimGeneral(Syntax::dMDC);
  }
  delete event;
}

void CopyEventHandler::markedSectionEnd(MarkedSectionEndEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event)) {
    inSpecialMarkedSection_ = 0;
    return;
  }
  if (!(normalizeFlags_ & normalizeMarkedSection)
      || (inInstance_ && inSpecialMarkedSection_))
    outputMarkup(event->location(), event->markup());
  else if (inInstance_) {
    // Put an empty comment so that REs aren't changed.
    handleChange();
    os() << syntax_->delimGeneral(Syntax::dMDO)
         << syntax_->delimGeneral(Syntax::dMDC);
  }
  inSpecialMarkedSection_ = 0;
  delete event;
}

void CopyEventHandler::ignoredChars(IgnoredCharsEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  if (!(normalizeFlags_ & normalizeMarkedSection))
    os().write(event->data(), event->dataLength());
  delete event;
}

void CopyEventHandler::usemap(UsemapEvent *event)
{
  omittagHoist_ = 0;
  if (doNothing(event))
    return;
  if (!(normalizeFlags_ & normalizeShortref))
    outputMarkup(event->location(), event->markup());
  else if (inInstance_) {
    // Put an empty comment so that REs aren't changed.
    handleChange();
    os() << syntax_->delimGeneral(Syntax::dMDO)
         << syntax_->delimGeneral(Syntax::dMDC);
  }
  delete event;
}

void CopyEventHandler::uselink(UselinkEvent *event)
{
  omittagHoist_ = 0;
  markup(event->location(), event->markup());
  delete event;
}

void CopyEventHandler::startDtd(StartDtdEvent *event)
{
  startSubset(event);
}

void CopyEventHandler::startLpd(StartLpdEvent *event)
{
  startSubset(event);
}

void CopyEventHandler::startSubset(StartSubsetEvent *event)
{
  if (doNothing(event))
    return;
  if (!event->entity().isNull()
      && (normalizeFlags_ & normalizeExpandProlog)) {
    const Markup &m = event->markup();
    for (MarkupIter iter(m); iter.valid(); iter.advance())
      if (iter.type() == Markup::reservedName
	  && (iter.reservedName() == Syntax::rSYSTEM
	      || iter.reservedName() == Syntax::rPUBLIC)) {
	Markup copy(m);
	copy.setLength(iter.index());
	outputMarkup(event->location(), copy);
	break;
      }
  }
  else
    outputMarkup(event->location(), event->markup());
  if (event->hasInternalSubset()
      || (normalizeFlags_ & normalizeExpandProlog)) {
    os() << syntax_->delimGeneral(Syntax::dDSO);
    hasInternalSubset_ = 1;
  }
  else
    hasInternalSubset_ = 0;
  delete event;
}

void CopyEventHandler::endDtd(EndDtdEvent *event)
{
  endSubset(event);
}

void CopyEventHandler::endLpd(EndLpdEvent *event)
{
  endSubset(event);
}

void CopyEventHandler::endSubset(MarkupEvent *event)
{
  if (doNothing(event))
    return;
  if (hasInternalSubset_)
    os() << syntax_->delimGeneral(Syntax::dDSC);
  outputMarkup(event->location(), event->markup());
  delete event;
}

void CopyEventHandler::entityDecl(EntityDeclEvent *event)
{
  currentAttributes_ = 0;
  const ExternalDataEntity *extData = event->entity().asExternalDataEntity();
  if (extData)
    currentAttributes_ = &extData->attributes();
  markup(event->location(), event->markup());
  currentAttributes_ = 0;
  delete event;
}

void CopyEventHandler::shortrefDecl(ShortrefDeclEvent *event)
{
  if (doNothing(event))
    return;
  if (!(normalizeFlags_ & normalizeShortref))
    outputMarkup(event->location(), event->markup());
  delete event;
}

void CopyEventHandler::entityStart(EntityStartEvent *event)
{
  if (event->entity()->name() == outputEntity_
      && event->entity()->declType() == Entity::generalEntity)
    outputEntityLevel_ = entityLevel_ + 1;
  if (inInstance_ && (normalizeFlags_ & normalizeOmittagHoist)) {
    if (event->entity()->asInternalEntity())
      omittagHoist_++;
    else
      omittagHoist_ = 0;
  }
  if (doNothing(event)) {
    entityLevel_++;
    return;
  }
  entityLevel_++;
  if ((normalizeFlags_ & normalizeExpand)
      && inInstance_
      && entityLevel_ > outputEntityLevel_) {
    entityStack_.grow().ref = event->entityOrigin();
    os_ = &entityStack_.last().str;
  }
  entityOrigin_ = event->entityOrigin();
  delete event;
}

void CopyEventHandler::entityEnd(EntityEndEvent *event)
{
  if (omittagHoist_ > 0)
    omittagHoist_--;
  if (entityLevel_-- == outputEntityLevel_) {
    outputEntityLevel_ = unsigned(-1);
    outputEntity_.clear();
  }
  else if (!(normalizeFlags_
	     & (inInstance_ ? normalizeExpand : normalizeExpandProlog))
	   && entityLevel_ == outputEntityLevel_) {
    if (!entityOrigin_.isNull()) {
      switch (entityOrigin_->entity()->declType()) {
      case Entity::doctype:
      case Entity::linktype:
	break;
      default:
	entityRef(entityOrigin_.pointer());
	break;
      }
    }
    entityOrigin_.clear();
  }
  else if ((normalizeFlags_ & normalizeExpand)
	   && inInstance_
	   && entityLevel_ >= outputEntityLevel_) {
    if (entityStack_.length() > 0) {
      ConstResourcePointer<EntityOrigin> origin
	= entityStack_.last().ref;
      entityStack_.setLength(entityStack_.length() - 1);
      if (entityStack_.length() > 0)
	os_ = &entityStack_.last().str;
      else
	os_ = topOs_;
      entityRef(origin.pointer());
    }
  }
  delete event;
}

void CopyEventHandler::outputMarkup(const Location &loc,
				    const Markup &markup)
{
  int level = 0;
  Boolean first = 1;
  MarkupIter iter(markup);
  while (iter.valid()) {
    switch (iter.type()) {
    case Markup::delimiter:
      if (first)
	withNamedCharRef(syntax_->delimGeneral(iter.delimGeneral()), loc);
      else if (!level) {
	os() << syntax_->delimGeneral(iter.delimGeneral());
	// hack, hack!
	if (iter.delimGeneral() == Syntax::dDSO && currentAttributes_ != 0) {
	  attributeSpecList(iter, *currentAttributes_);
	  first = 0;
	  continue;		// skip the advance
	}
      }
      break;
    case Markup::refEndRe:
      if (!level)
	os().put(syntax_->standardFunction(Syntax::fRE));
      break;
    case Markup::sdReservedName:
      if (!level) {
	if (normalizeFlags_ & normalizeReserved)
	  os() << sd_->reservedName(iter.sdReservedName());
	else
	  os() << iter;
      }
      break;
    case Markup::reservedName:
      if (!level && (normalizeFlags_ & normalizeReserved)) {
	os() << syntax_->reservedName(iter.reservedName());
	break;
      }
    case Markup::shortref:
      if (first) {
	withNamedCharRef(iter.charsPointer(), iter.charsLength(), loc);
	break;
      }
      // fall through
    case Markup::name:
    case Markup::nameToken:
    case Markup::attributeValue:
    case Markup::number:
    case Markup::s:
      if (!level)
	os() << iter;
      break;
    case Markup::comment:
      if (!level)
	os() << syntax_->delimGeneral(Syntax::dCOM)
	     << iter
	     << syntax_->delimGeneral(Syntax::dCOM);
      break;
    case Markup::entityStart:
      if (!level++) {
	const EntityOrigin *origin = iter.entityOrigin();
	// entityStarts in the SGML declaration don't have explicit references
	if (!origin->entity().isNull())
	  entityRef(origin);
      }
      break;
    case Markup::entityEnd:
      level--;
      break;
    case Markup::literal:
      if (!level)
	literal(iter.text());
      break;
    case Markup::sdLiteral:
      if (!level)
	sdParamLiteral(iter.sdText());
      break;
    default:
      CANNOT_HAPPEN();
    }
    iter.advance();
    first = 0;
  }
}

void CopyEventHandler::sdParamLiteral(const SdText &text)
{
  const CString &delim = syntax_->delimGeneral(text.lita()
					       ? Syntax::dLITA
					       : Syntax::dLIT);
  os() << delim;
  SdTextIter iter(text);
  const SyntaxChar *p;
  size_t n;
  Location loc;
  while (iter.next(p, n, loc)) {
    const Markup *markupPtr;
    if (n == 1 && loc.origin()->isNumericCharRef(markupPtr)) {
      if (markupPtr)
	outputMarkup(loc.origin()->parent(), *markupPtr);
    }
    else if (n > 0) {
      Char c = Char(*p);
      withNamedCharRef(&c, 1, loc);
      for (++p, --n; n > 0; ++p, --n)
	os().put(Char(*p));
    }
  }
  os() << delim;
}

void CopyEventHandler::entityRef(const EntityOrigin *origin)
{
  const Markup *m = origin->markup();
  if (!m)
    return;
  MarkupIter iter(*m);
  if (iter.valid()) {
    iter.advance();
    if (iter.valid()
	&& iter.type() == Markup::shortref
	&& (normalizeFlags_ & normalizeShortref)) {
       handleChange();
       Boolean containsRE = 0;
       Boolean containsRS = 0;
       for (size_t i = 0; i < iter.charsLength(); i++) {
	 Char c = iter.charsPointer()[i];
	 if (c == syntax_->standardFunction(Syntax::fRE))
	   containsRE = 1;
	 else if (c == syntax_->standardFunction(Syntax::fRS))
	   containsRS = 1;
       }
       if (containsRS)
	 os().put(syntax_->standardFunction(Syntax::fRS));
       os() << syntax_->delimGeneral(Syntax::dERO)
	 << origin->entity()->name();
       if (containsRE)
	 os().put(syntax_->standardFunction(Syntax::fRE));
       else
	 os() << syntax_->delimGeneral(Syntax::dREFC);
       return;
     }
  }
  outputMarkup(origin->parent(), *m);
}

const CString &CopyEventHandler::elementTypeOrigName(const ElementType *type,
						     CString &buf)
{
  if (type->index() < elementTypeOrigNames_.length()
      && elementTypeOrigNames_[type->index()].length() > 0)
    return elementTypeOrigNames_[type->index()];
  else
    return generalName(type->name(), buf);
}

const CString &CopyEventHandler::generalName(const CString &name,
					     CString &buf)
{
  if ((normalizeFlags_ & normalizeLower)
      && syntax_->namecaseGeneral())
    return lowerCaseName(name, buf);
  else
    return name;
}

const CString &CopyEventHandler::entityName(const CString &name,
					    CString &buf)
{
  if ((normalizeFlags_ & normalizeLower)
      && syntax_->namecaseEntity())
    return lowerCaseName(name, buf);
  else
    return name;
}

const CString &CopyEventHandler::lowerCaseName(const CString &name,
					       CString &buf)
{
  size_t i;
  for (i = 0; i < name.length(); i++) {
    Char c = lowerSubst_[name[i]];
    if (c != name[i]) {
      buf = name;
      buf[i] = c;
      for (i++; i < name.length(); i++)
	lowerSubst_.subst(buf[i]);
      return buf;
    }
  }
  return name;
}

void CopyEventHandler::handleChange()
{
  if (os_ != topOs_) {
    os_ = topOs_;
    for (size_t i = 0; i < entityStack_.length(); i++) {
      CString tem;
      entityStack_[i].str.flush();
      entityStack_[i].str.extractString(tem);
      os() << tem;
    }
    entityStack_.setLength(0);
  }
}
