// Copyright (c) 1995 James Clark
// See the file COPYING for copying permission.

#include "config.h"
#include "SgmlParser.H"
#include "Event.H"
#include "MessageReporter.H"
#include "OutputCharStream.H"
#include "ParserToolkit.H"
#include "Boolean.H"
#include "programName.H"
#include "macros.H"
#include "getopt.h"
#include "SpamMessages.H"
#include "LineOutputCodingSystem.H"
#include "CopyEventHandler.H"
#include "InputSource.H"
#include "ErrnoMessageArg.H"

#include "outOfMemory.H"
#include "xnew.h"

#include <stdlib.h>
#include <fstream.h>
#include <iostream.h>
#include <errno.h>
#ifdef HAVE_LOCALE
#include <locale.h>
#endif

class XCopyEventHandler : public CopyEventHandler {
public:
  XCopyEventHandler(OutputCharStream *os, unsigned normalizeFlags,
		    const CString &outputEntity, MessageReporter *);
  Boolean hadError() const;
  void message(MessageEvent *);
private:
  void reportTagInSpecialMarkedSection(const Location &);
  MessageReporter *reporter_;
  OutputCharStream *os_;
  Boolean hadError_;
};

inline
Boolean XCopyEventHandler::hadError() const
{
  return hadError_;
}

inline
void usage(Messenger &mgr)
{
  mgr.message(SpamMessages::usageInfo);
}

const Char RE = '\r';
const Char RS = '\n';

static
struct {
  const char *name;
  unsigned flag;
} normalizeTable[] = {
  { "omittag", CopyEventHandler::normalizeOmittag },
  { "shortref", CopyEventHandler::normalizeShortref },
  { "net", CopyEventHandler::normalizeNet },
  { "emptytag", CopyEventHandler::normalizeEmptytag },
  { "unclosed", CopyEventHandler::normalizeUnclosed },
  { "attname", CopyEventHandler::normalizeAttname },
  { "attvalue", CopyEventHandler::normalizeAttvalue },
  { "attspec", CopyEventHandler::normalizeAttspec },
  { "current", CopyEventHandler::normalizeCurrent },
  { "shorttag", CopyEventHandler::normalizeShorttag },
  { "rank", CopyEventHandler::normalizeRank },
  { "reserved", CopyEventHandler::normalizeReserved },
  { "ms", CopyEventHandler::normalizeMarkedSection },
};

int main(int argc, char **argv)
{
  set_new_handler(outOfMemory);
#ifdef HAVE_LOCALE
  setlocale(LC_ALL, "");
#endif
#ifdef __GNUG__
  // cout is a performance disaster in libg++ unless we do this.
  ios::sync_with_stdio(0);
#endif
  programName = argv[0];
  int opt;
  char arg[2];
  ParserToolkit kit;
  unsigned normalizeFlags = 0;
  Boolean rawOutput = 0;
  CString outputEntity;
  const char *errorFile = 0;
  
  opterr = 0;			// no messages from getopt
  // Colon at beginning is Posix.2ism that says to return : rather than ? for
  // missing option argument.
  while ((opt = getopt(argc, argv, ":D:i:m:o:ef:hlpw:c:vxr")) >= 0) {
    switch (opt) {
    case 'c':
      kit.addCatalogSysid(optarg);
      break;
    case 'D':
      kit.addSearchDir(optarg);
      break;
    case 'e':
      // describe open entities in error messages
      kit.addOption(MessageReporter::openEntities);
      break;
    case 'f':
      errorFile = optarg;
      break;
    case 'h':
      // hoist omitted tags out of entities
      normalizeFlags |= CopyEventHandler::normalizeOmittagHoist;
      break;
    case 'i':
      // pretend that optarg is defined as INCLUDE
      kit.addInclude(optarg);
      break;
    case 'l':			// prefer lowercase
      normalizeFlags |= CopyEventHandler::normalizeLower;
      break;
    case 'm':
      {
	Boolean found = 0;
	for (size_t i = 0; i < SIZEOF(normalizeTable); i++)
	  if (strcmp(normalizeTable[i].name, optarg) == 0) {
	    found = 1;
	    normalizeFlags |= normalizeTable[i].flag;
	    break;
	  }
	if (!found) {
	  kit.message(SpamMessages::invalidMarkupArgument,
		       StringMessageArg(kit.codingSystem()->convertIn(optarg)));
	  usage(kit);
	}
      }
      break;
    case 'o':
      outputEntity = kit.codingSystem()->convertIn(optarg);
      break;
    case 'p':
      if (normalizeFlags & CopyEventHandler::normalizeIncludeProlog)
	normalizeFlags |= CopyEventHandler::normalizeExpandProlog;
      else
	normalizeFlags |= CopyEventHandler::normalizeIncludeProlog;
      break;
    case 'r':
      rawOutput = 1;
      break;
    case 'v':
      // print the version number
      {
	extern const char *versionString;
	// FIXME perhaps we should have a constant argument instead
	kit.message(SpamMessages::versionInfo,
		     StringMessageArg(kit.codingSystem()
				      ->convertIn(versionString)));
      }
      break;
    case 'w':
      if (!kit.enableWarning(optarg))
	kit.message(SpamMessages::unknownWarning,
		     StringMessageArg(kit.codingSystem()
				      ->convertIn(optarg)));
      break;
    case 'x':
      if (normalizeFlags & CopyEventHandler::normalizeExpand)
	normalizeFlags |= CopyEventHandler::normalizeExpandAll;
      else
	normalizeFlags |= CopyEventHandler::normalizeExpand;
      break;
    case ':':
      arg[0] = optopt;
      arg[1] = '\0';
      kit.message(SpamMessages::missingOptionArgError,
		   StringMessageArg(kit.codingSystem()->convertIn(arg)));
      usage(kit);
      return 1;
    case '?':
      arg[0] = optopt;
      arg[1] = '\0';
      kit.message(SpamMessages::invalidOptionError,
		   StringMessageArg(kit.codingSystem()->convertIn(arg)));
      usage(kit);
      return 1;
    default:
      CANNOT_HAPPEN();
    }
  }
  if (errorFile) {
    static filebuf file;
    if (!file.open(errorFile, ios::out|ios::trunc)) {
      kit.message(SpamMessages::cannotOpenOutputError,
		   StringMessageArg(kit.codingSystem()
				    ->convertIn(errorFile)),
		   ErrnoMessageArg(errno));
      exit(1);
    }
    kit.setErrorStream(&file);
  }
  kit.options.eventsWanted.addInstanceMarkup();
  kit.options.eventsWanted.addPrologMarkup();
  if (!kit.init(argc - optind, argv + optind))
    return 1;
  SgmlParser &parser = kit.parser();
  LineOutputCodingSystem line(kit.codingSystem(), RS, RE, '\n');
  const OutputCodingSystem *ocs
    = (rawOutput
       ? (const OutputCodingSystem *)kit.codingSystem()
       : (const OutputCodingSystem *)&line);
  IosOutputCharStream os(cout.rdbuf(), ocs);
  XCopyEventHandler handler(&os, normalizeFlags, outputEntity, &kit);
  parser.parseAll(handler);
  return handler.hadError();
}

XCopyEventHandler::XCopyEventHandler(OutputCharStream *os,
				     unsigned normalizeFlags,
				     const CString &outputEntity,
				     MessageReporter *reporter)
: reporter_(reporter), hadError_(0), os_(os),
  CopyEventHandler(os, normalizeFlags, outputEntity)
{
}

void XCopyEventHandler::message(MessageEvent *event)
{
  reporter_->reportMessage(event->message());
  if (event->message().isError())
    hadError_ = 1;
  delete event;
}

void XCopyEventHandler::reportTagInSpecialMarkedSection(const Location &loc)
{
  reporter_->setNextLocation(loc);
  reporter_->message(SpamMessages::tagInSpecialMarkedSection);
}

