// Copyright (c) 1994, 1995 James Clark
// See the file COPYING for copying permission.

#ifdef __GNUG__
#pragma implementation
#endif

#include "config.h"
#include "CharsetInfo.H"
#include "MessageArg.H"
#include "CatalogMessages.H"
#include "SOEntityCatalog.H"
#include "EntityDecl.H"
#include "Message.H"
#include "CString.H"
#include "types.H"
#include "HashTable.H"
#include "InputSource.H"
#include "Boolean.H"
#include "SubstTable.H"
#include "CatalogEntry.H"

class SOEntityCatalogImpl : public SOEntityCatalog {
public:
  SOEntityCatalogImpl(ConstResourcePointer<ExtendEntityManager> em);
  typedef EntityDecl::DeclType DeclType;
  Boolean document(const CharsetInfo &, Messenger &, CString &) const;
  Boolean sgmlDecl(const CharsetInfo &, Messenger &, CString &) const;
  Boolean lookup(const EntityDecl &entity,
		 const Syntax &,
		 const CharsetInfo &,
		 Messenger &,
		 CString &) const;
  Boolean lookupPublic(const CString &,
		       const CharsetInfo &,
		       Messenger &,
		       CString &) const;
  SOEntityCatalogImpl();
  void addPublicId(CString &publicId, CString &systemId, const Location &,
		   Boolean override);
  void addSystemId(CString &systemId, CString &replSystemId, const Location &);
  void addName(CString &name, DeclType, CString &systemId, const Location &,
	       Boolean override);
  void setSgmlDecl(CString &str, const Location &loc);
  void setDocument(CString &str, const Location &loc);
  void endCatalog();
private:
  SOEntityCatalogImpl(const SOEntityCatalogImpl &);	// undefined
  void operator=(const SOEntityCatalogImpl &); // undefined

  class Table {
  public:
    Table();
    const CatalogEntry *lookup(const CString &, Boolean overrideOnly) const;
    const CatalogEntry *lookup(const CString &key,
			       const SubstTable<Char> &substTable,
			       Boolean overrideOnly) const;
    void insert(const CString &, const CatalogEntry &, Boolean override);
    size_t count() const;
  private:
    Table(const Table &);	// undefined
    void operator=(const Table &); // undefined
    // These are entries that are applicable when an explicit system id
    // was specified in the external identifier.
    HashTable<CString,CatalogEntry> overrideEntries_;
    // This specifies the entries that should substitute for the
    // overrideEntries_ when an explicit system identifier was not specified.
    HashTable<CString,CatalogEntry> normalEntries_;
  };

  Table publicIds_;
  HashTable<CString,CatalogEntry> systemIds_;
  Table names_[4];
  size_t catalogNumber_;
  Boolean haveSgmlDecl_;
  CString sgmlDecl_;
  Location sgmlDeclLoc_;
  CString document_;
  Boolean haveDocument_;
  Location documentLoc_;
  ConstResourcePointer<ExtendEntityManager> em_;
};

class CatalogParser : private Messenger {
public:
  CatalogParser(const CharsetInfo &);
  void parseCatalog(const CString &sysid,
		    Boolean mustExist,
		    const CharsetInfo &sysidCharset,
		    const CharsetInfo &catalogCharset,
		    InputSourceOrigin *origin,
		    const ConstResourcePointer<ExtendEntityManager> &em,
		    SOEntityCatalogImpl *catalog,
		    Messenger &mgr);
public:
  // Since it's a return type, it has to be public to keep some 
  // (broken) compilers happy.
  enum Param {
    eofParam,
    literalParam,
    nameParam,
    percentParam
  };
private:
  enum {
    data,
    eof,
    nul,
    lit,
    lita,
    minus,
    s,
    min				// other minimum data characters
  };
  enum { minimumLiteral = 01 };

  Messenger &messenger() { return *this; }
  void dispatchMessage(Message &);
  void initMessage(Message &);
  void parsePublic();
  void parseSystem();
  void parseNameMap(EntityDecl::DeclType declType);
  void parseOverride();
  Param parseParam(unsigned flags = 0);
  Boolean parseArg();
  void parseLiteral(Char delim, unsigned flags);
  void parseName();
  void skipComment();
  void upcase(CString &);
  Boolean inLoop(const Location &loc);
  Boolean isMinimumData(Xchar c) {
    int cat = categoryTable_[c];
    return (cat == min || (cat == s && c != tab_)
	    || cat == minus || cat == lita);
  }
  Xchar get() { return in_->get(messenger()); }
  void unget() { in_->ungetToken(); }
  Messenger *mgr_;
  InputSource *in_;
  SOEntityCatalogImpl *catalog_;
  CString param_;
  Location paramLoc_;
  Char minus_;
  Char tab_;
  Char rs_;
  Char re_;
  Char space_;
  CString publicKey_;
  CString systemKey_;
  CString entityKey_;
  CString doctypeKey_;
  CString linktypeKey_;
  CString notationKey_;
  CString overrideKey_;
  CString sgmlDeclKey_;
  CString documentKey_;
  CString catalogKey_;
  CString yesKey_;
  CString noKey_;
  XcharMap<unsigned char> categoryTable_;
  SubstTable<Char> substTable_;
  Boolean override_;
};

SOEntityCatalog *SOEntityCatalog::make(const CString *sysids,
				       size_t nSysids,
				       size_t nSysidsMustExist,
				       const CharsetInfo &sysidCharset,
				       const CharsetInfo &catalogCharset,
				       ConstResourcePointer<ExtendEntityManager> em,
				       Messenger &mgr)
{
  SOEntityCatalogImpl *impl = new SOEntityCatalogImpl(em);
  CatalogParser parser(catalogCharset);
  for (size_t i = 0; i < nSysids; i++)
    parser.parseCatalog(sysids[i], i < nSysidsMustExist, sysidCharset,
			catalogCharset, new InputSourceOrigin, em, impl, mgr);
  return impl;
}

SOEntityCatalogImpl::SOEntityCatalogImpl(ConstResourcePointer<ExtendEntityManager> em)
: em_(em), catalogNumber_(0), haveSgmlDecl_(0), haveDocument_(0)
{
}

void SOEntityCatalogImpl::endCatalog()
{
  catalogNumber_++;
}

Boolean SOEntityCatalogImpl::lookup(const EntityDecl &entity,
				    const Syntax &syntax,
				    const CharsetInfo &charset,
				    Messenger &mgr,
				    CString &result) const
{
  const CatalogEntry *entry = 0;
  if (entity.systemIdPointer())
    entry = systemIds_.lookup(*entity.systemIdPointer());
  if (entity.publicIdPointer()) {
    const CatalogEntry *publicEntry;
    publicEntry = publicIds_.lookup(*entity.publicIdPointer(),
				    entity.systemIdPointer() != 0);
    // match for system id has priority over match for public id in same
    // catalog
    if (publicEntry
	&& (!entry || publicEntry->catalogNumber < entry->catalogNumber))
      entry = publicEntry;
  }
  if (entity.name().length() > 0
      && (!entry || entry->catalogNumber > 0)) {
    const CatalogEntry *entityEntry; 
    int tableIndex = (entity.declType() >= EntityDecl::parameterEntity
		      ? int(entity.declType()) - 1
		      : int(entity.declType()));
    CString name(entity.name());
    Boolean subst;
    switch (entity.declType()) {
    case EntityDecl::parameterEntity:
      {
	CString tem(name);
	name = syntax.peroDelim();
	tem += name;
      }
      // fall through
    case EntityDecl::generalEntity:
      subst = syntax.namecaseEntity();
      break;
    default:
      subst = syntax.namecaseGeneral();
      break;
    }
    if (!subst)
      entityEntry = names_[tableIndex].lookup(name,
					      entity.systemIdPointer() != 0);
    else
      entityEntry = names_[tableIndex].lookup(entity.name(),
					      syntax.upperSubstTable(),
					      entity.systemIdPointer() != 0);
    // match for public id has priority over match for entity in same
    // catalog
    if (entityEntry
	&& (!entry || entityEntry->catalogNumber < entry->catalogNumber))
      entry = entityEntry;
  }
  if (entry)
    return em_->expandSystemId(entry->to,
			       entry->loc,
			       entity.dataType() == EntityDecl::ndata,
			       charset,
			       mgr,
			       result);
  if (entity.systemIdPointer())
    return em_->expandSystemId(*entity.systemIdPointer(),
			       entity.defLocation(),
			       entity.dataType() == EntityDecl::ndata,
			       charset,
			       mgr,
			       result);
  return 0;
}

Boolean SOEntityCatalogImpl::lookupPublic(const CString &publicId,
					  const CharsetInfo &charset,
					  Messenger &mgr,
					  CString &result) const
{
  const CatalogEntry *entry = publicIds_.lookup(publicId, 0);
  return (entry
	  && em_->expandSystemId(entry->to, entry->loc, 0, charset, mgr,
				 result));
				 
}

Boolean SOEntityCatalogImpl::sgmlDecl(const CharsetInfo &charset,
				      Messenger &mgr,
				      CString &result) const

{
  return haveSgmlDecl_ && em_->expandSystemId(sgmlDecl_, sgmlDeclLoc_,
					      0, charset, mgr, result);
}

Boolean SOEntityCatalogImpl::document(const CharsetInfo &charset,
				      Messenger &mgr,
				      CString &result) const

{
  return haveDocument_ && em_->expandSystemId(document_, documentLoc_,
					      0, charset, mgr, result);
}

void SOEntityCatalogImpl::addPublicId(CString &publicId, CString &systemId,
				      const Location &loc, Boolean override)
{
  CatalogEntry entry;
  entry.loc = loc;
  entry.catalogNumber = catalogNumber_;
  systemId.moveTo(entry.to);
  publicIds_.insert(publicId, entry, override);
}

void SOEntityCatalogImpl::addSystemId(CString &systemId, CString &toSystemId,
				      const Location &loc)
{
  CatalogEntry entry;
  entry.loc = loc;
  entry.catalogNumber = catalogNumber_;
  toSystemId.moveTo(entry.to);
  systemIds_.insert(systemId, entry, false);
}

void SOEntityCatalogImpl::addName(CString &name, DeclType declType,
				  CString &systemId, const Location &loc,
				  Boolean override)
{
  CatalogEntry entry;
  entry.loc = loc;
  entry.catalogNumber = catalogNumber_;
  int tableIndex = (declType >= EntityDecl::parameterEntity
		    ? int(declType) - 1
		    : int(declType));
  entry.serial = names_[tableIndex].count();
  systemId.moveTo(entry.to);
  names_[tableIndex].insert(name, entry, override);
}

void SOEntityCatalogImpl::setSgmlDecl(CString &str, const Location &loc)
{
  if (!haveSgmlDecl_) {
    haveSgmlDecl_ = true;
    str.moveTo(sgmlDecl_);
    sgmlDeclLoc_ = loc;
  }
}

void SOEntityCatalogImpl::setDocument(CString &str, const Location &loc)
{
  if (!haveDocument_) {
    haveDocument_ = true;
    str.moveTo(document_);
    documentLoc_ = loc;
  }
}

SOEntityCatalogImpl::Table::Table()
{
}

void SOEntityCatalogImpl::Table::insert(const CString &key, const CatalogEntry &entry,
			    Boolean override)
{
  if (override)
    overrideEntries_.insert(key, entry, false);
  else {
    const CatalogEntry *e = overrideEntries_.lookup(key);
    if (!e)
      normalEntries_.insert(key, entry, false);
  }
}

const CatalogEntry *SOEntityCatalogImpl::Table::lookup(const CString &key,
					   Boolean overrideOnly) const
{
  if (!overrideOnly) {
    const CatalogEntry *e = normalEntries_.lookup(key);
    if (e)
      return e;
  }
  return overrideEntries_.lookup(key);
}

const CatalogEntry *SOEntityCatalogImpl::Table::lookup(const CString &name,
					   const SubstTable<Char> &substTable,
					   Boolean overrideOnly) const
{
  HashTableIter<CString,CatalogEntry> iter1(overrideEntries_);
  HashTableIter<CString,CatalogEntry> iter2(normalEntries_);
  HashTableIter<CString,CatalogEntry> *iters[2];
  int nIter = 0;
  iters[nIter++] = &iter1;
  if (!overrideOnly)
    iters[nIter++] = &iter2;
  const CatalogEntry *entry = 0;
  for (int i = 0; i < nIter; i++) {
    HashTableIter<CString,CatalogEntry> &iter = *iters[i];
    const CString *key;
    const CatalogEntry *value;
    CString buffer;
    while (iter.next(key, value)) {
      buffer = *key;
      for (size_t j = 0; j < buffer.length(); j++)
	substTable.subst(buffer[j]);
      if (buffer == name) {
	if (!entry || value->serial < entry->serial)
	  entry = value;
      }
    }
  }
  return entry;
}

size_t SOEntityCatalogImpl::Table::count() const
{
  return normalEntries_.count() + overrideEntries_.count();
}

CatalogParser::CatalogParser(const CharsetInfo &charset)
: categoryTable_(data),
  entityKey_(charset.execToDesc("ENTITY")),
  publicKey_(charset.execToDesc("PUBLIC")),
  systemKey_(charset.execToDesc("SYSTEM")),
  doctypeKey_(charset.execToDesc("DOCTYPE")),
  linktypeKey_(charset.execToDesc("LINKTYPE")),
  notationKey_(charset.execToDesc("NOTATION")),
  overrideKey_(charset.execToDesc("OVERRIDE")),
  sgmlDeclKey_(charset.execToDesc("SGMLDECL")),
  documentKey_(charset.execToDesc("DOCUMENT")),
  catalogKey_(charset.execToDesc("CATALOG")),
  yesKey_(charset.execToDesc("YES")),
  noKey_(charset.execToDesc("NO"))
{
  static const char lcletters[] = "abcdefghijklmnopqrstuvwxyz";
  static const char ucletters[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  // minimum data other than lcletter, ucletter
  static const char minChars[] = "0123456789-.'()+,/:=?";
  static const char sChars[] = " \n\r\t";
  categoryTable_.setChar(0, nul);
  const char *p;
  const char *q;
  for (p = lcletters, q = ucletters; *p; p++, q++) {
    Char lc = charset.execToDesc(*p);
    Char uc = charset.execToDesc(*q);
    substTable_.addSubst(lc, uc);
    categoryTable_.setChar(lc, min);
    categoryTable_.setChar(uc, min);
  }
  for (p = sChars; *p; p++)
    categoryTable_.setChar(charset.execToDesc(*p), s);
  for (p = minChars; *p; p++)
    categoryTable_.setChar(charset.execToDesc(*p), min);
  categoryTable_.setChar(charset.execToDesc('\''), lita);
  categoryTable_.setChar(charset.execToDesc('"'), lit);
  minus_ = charset.execToDesc('-');
  categoryTable_.setChar(minus_, minus);
  tab_ = charset.execToDesc('\t');
  re_ = charset.execToDesc('\r');
  rs_ = charset.execToDesc('\n');
  space_ = charset.execToDesc(' ');
  categoryTable_.setEe(eof);
}

void CatalogParser::parseCatalog(const CString &sysid,
				 Boolean mustExist,
				 const CharsetInfo &sysidCharset,
				 const CharsetInfo &catalogCharset,
				 InputSourceOrigin *origin,
				 const ConstResourcePointer<ExtendEntityManager> &em,
				 SOEntityCatalogImpl *catalog,
				 Messenger &mgr)
{
  in_ = (mustExist
	 ? em->open(sysid, sysidCharset, origin, 0, mgr)
	 : em->openIfExists(sysid, sysidCharset, origin, 0, mgr));
  if (!in_)
    return;
  catalog_ = catalog;
  mgr_ = &mgr;
  override_ = 0;
  Boolean recovering = false;
  GrowableVectorD<CString> subSysids;
  GrowableVectorD<Location> subSysidLocs;
  for (;;) {
    Param parm = parseParam();
    if (parm == nameParam) {
      upcase(param_);
      recovering = false;
      if (param_ == publicKey_)
	parsePublic();
      else if (param_ == systemKey_)
	parseSystem();
      else if (param_ == entityKey_)
	parseNameMap(EntityDecl::generalEntity);
      else if (param_ == doctypeKey_)
	parseNameMap(EntityDecl::doctype);
      else if (param_ == linktypeKey_)
	parseNameMap(EntityDecl::linktype);
      else if (param_ == notationKey_)
	parseNameMap(EntityDecl::notation);
      else if (param_ == sgmlDeclKey_) {
	if (parseArg())
	  catalog_->setSgmlDecl(param_, paramLoc_);
      }
      else if (param_ == documentKey_) {
	if (parseArg())
	  catalog_->setDocument(param_, paramLoc_);
      }
      else if (param_ == overrideKey_)
	parseOverride();
      else if (param_ == catalogKey_) {
	if (parseArg()) {
	  if (inLoop(paramLoc_))
	    break;
	  param_.moveTo(subSysids.grow());
	  subSysidLocs.grow() = paramLoc_;
	}
      }
      else
	recovering = true;
    }
    else if (parm == eofParam)
      break;
    else if (!recovering) {
      recovering = true;
      message(CatalogMessages::nameExpected);
      if (parseParam() == eofParam)
	break;
    }
  }
  delete in_;
  catalog->endCatalog();
  for (size_t i = 0; i < subSysids.length(); i++) {
    CString tem;
    if (em->expandSystemId(subSysids[i], subSysidLocs[i], 0, catalogCharset,
			   mgr, tem))
      parseCatalog(tem, 1, catalogCharset, catalogCharset,
		   new InputSourceOrigin(subSysidLocs[i]), em, catalog, mgr);
  }
}

Boolean CatalogParser::inLoop(const Location &loc)
{
  const InputSourceOrigin *origin = paramLoc_.origin()->asInputSourceOrigin();
  if (!origin)
    return 0;
  const ExternalInfo *info = origin->externalInfo();
  if (!info)
    return 0;
  StorageObjectLocation soLoc;
  if (!ExtendEntityManager::externalize(info,
					origin->startOffset(paramLoc_.index()),
					soLoc))
    return 0;
  for (;;) {
    const Location &parent = origin->parent();
    if (parent.origin().isNull())
      break;
    origin = parent.origin()->asInputSourceOrigin();
    if (!origin)
      break;
    const ExternalInfo *info1 = origin->externalInfo();
    if (info1) {
      StorageObjectLocation soLoc1;
      if (ExtendEntityManager::externalize(info1,
					   origin->startOffset(parent.index()),
					   soLoc1)) {
	const StorageObjectSpec *sos = soLoc.storageObjectSpec;
	const StorageObjectSpec *sos1 = soLoc1.storageObjectSpec;
	if (sos->storageManager == sos1->storageManager
	    && sos->id == sos1->id) {
	  setNextLocation(loc.origin()->parent());
	  message(CatalogMessages::inLoop);
	  return 1;
	}
      }
    }
  }
  return 0;
}

void CatalogParser::parseOverride()
{
  if (parseParam() != nameParam) {
    message(CatalogMessages::overrideYesOrNo);
    return;
  }
  upcase(param_);
  if (param_ == yesKey_)
    override_ = 1;
  else if (param_ == noKey_)
    override_ = 0;
  else
    message(CatalogMessages::overrideYesOrNo);
}

void CatalogParser::parsePublic()
{
  if (parseParam(minimumLiteral) != literalParam) {
    message(CatalogMessages::literalExpected);
    return;
  }
  CString publicId;
  param_.moveTo(publicId);
  if (!parseArg())
    return;
  catalog_->addPublicId(publicId, param_, paramLoc_, override_);
}

void CatalogParser::parseSystem()
{
  if (!parseArg())
    return;
  CString systemId;
  param_.moveTo(systemId);
  if (!parseArg())
    return;
  catalog_->addSystemId(systemId, param_, paramLoc_);
}

void CatalogParser::parseNameMap(EntityDecl::DeclType declType)
{
  if (!parseArg())
    return;
  CString name;
  param_.moveTo(name);
  if (!parseArg())
    return;
  catalog_->addName(name, declType, param_, paramLoc_, override_);
}

Boolean CatalogParser::parseArg()
{
  Param parm = parseParam();
  if (parm != nameParam && parm != literalParam) {
    message(CatalogMessages::nameOrLiteralExpected);
    return false;
  }
  return true;
}

CatalogParser::Param CatalogParser::parseParam(unsigned flags)
{
  for (;;) {
    Xchar c = get();
    switch (categoryTable_[c]) {
    case eof:
      return eofParam;
    case lit:
    case lita:
      parseLiteral(c, flags);
      return literalParam;
    case s:
      break;
    case nul:
      message(CatalogMessages::nulChar);
      break;
    case minus:
      c = get();
      if (c == minus_) {
	skipComment();
	break;
      }
      unget();
      // fall through
    default:
      parseName();
      return nameParam;
    }
  }
}

void CatalogParser::skipComment()
{
  for (;;) {
    Xchar c = get();
    if (c == minus_) {
      c = get();
      if (c == minus_)
	break;
    }
    if (c == InputSource::eE) {
      message(CatalogMessages::eofInComment);
      break;
    }
  }
}

void CatalogParser::parseLiteral(Char delim, unsigned flags)
{
  paramLoc_ = in_->currentLocation();
  enum { no, yesBegin, yesMiddle } skipping = yesBegin;
  param_.clear();
  for (;;) {
    Xchar c = get();
    if (c == InputSource::eE) {
      message(CatalogMessages::eofInLiteral);
      break;
    }
    if (Char(c) == delim)
      break;
    if (flags & minimumLiteral) {
      if (!isMinimumData(c))
	message(CatalogMessages::minimumData);
      if (c == rs_)
	;
      else if (c == space_ || c == re_) {
	if (skipping == no) {
	  param_ += space_;
	  skipping = yesMiddle;
	}
      }
      else {
	skipping = no;
	param_ += Char(c);
      }
    }
    else
      param_ += Char(c);
  }
  if (skipping == yesMiddle)
    param_.setLength(param_.length() - 1);
}

void CatalogParser::parseName()
{
  paramLoc_ = in_->currentLocation();
  size_t length;
  for (length = 1;; length++) {
    Xchar c = in_->tokenChar(messenger());
    int cat = categoryTable_[c];
    if (cat == eof || cat == s)
      break;
    // FIXME maybe check for LIT or LITA
    if (cat == nul)
      message(CatalogMessages::nulChar);
  }
  in_->endToken(length);
  param_.set(in_->currentTokenStart(), in_->currentTokenLength());
}

void CatalogParser::upcase(CString &str)
{
  for (size_t i = 0; i < str.length(); i++)
    substTable_.subst(str[i]);
}

void CatalogParser::dispatchMessage(Message &msg)
{
  mgr_->dispatchMessage(msg);
}

void CatalogParser::initMessage(Message &msg)
{
  msg.loc = in_->currentLocation();
}
